from typing import List, Mapping, Optional, Union

from kumoapi.jobs import BaselineJobRequest

from kumoai import global_state
from kumoai.client.jobs import BaselineJobID
from kumoai.graph import Graph
from kumoai.pquery.training_table import TrainingTable, TrainingTableJob
from kumoai.trainer.job import BaselineJob, BaselineJobResult


class BaselineTrainer:
    r"""A baseline trainer supports creating a Kumo baseline model on a
    :class:`~kumoai.pquery.PredictiveQuery`. It is primarily oriented around
    :meth:`~kumoai.trainer.Trainer.run`, which accepts a
    :class:`~kumoai.graph.Graph` and :class:`~kumoai.pquery.TrainingTable` and
    produces a :class:`~kumoai.trainer.BaselineJobResult`.

    Args:
        metrics List[str]: A list to metrics that baseline model will be
            evaluated on.

    Example:
        >>> import kumoai  # doctest: +SKIP
        >>> pquery = kumoai.PredictiveQuery(...)  # doctest: +SKIP
        >>> trainer = kumoai.BaselineTrainer(metrics=metrics)  # doctest: +SKIP

    .. # noqa: E501
    """
    def __init__(self, metrics: List[str]) -> None:
        self._metrics: List[str] = metrics

        # Cached from backend:
        self._baseline_job_id: Optional[BaselineJobID] = None

    def run(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
        *,
        non_blocking: bool = False,
        custom_tags: Mapping[str, str] = {},
    ) -> Union[BaselineJob, BaselineJobResult]:
        """Runs a baseline to the specified graph and training table.

        Args:
            graph (Graph): The :class:`~kumoai.graph.Graph` object that
                represents the tables and relationships that baseline model
                is running against.
            train_table (Union[TrainingTable, TrainingTableJob]): The
                :class:`~kumoai.pquery.TrainingTable`, or in-progress
                :class:`~kumoai.pquery.TrainingTableJob` that represents
                the training data produced by a
                :class:`~kumoai.pquery.PredictiveQuery` on :obj:`graph`.
            non_blocking (bool): Whether this operation should return
                immediately after launching the baseline job, or await
                completion of the baseline job. Defaults to False.
            custom_tags (Mapping[str, str], optional): Customer defined k-v
                tags to be associated with the job to be launched. Job tags
                are useful for grouping and searching jobs.. Defaults to {}.

        Returns:
            Union[BaselineJob, BaselineJobResult]:
                If ``non_blocking=False``, returns a baseline job object. If
                ``non_blocking=True``, returns a baseline job future object.
        """
        job_id = train_table.job_id
        assert job_id is not None

        train_table_job_api = global_state.client.generate_train_table_job_api
        pq_id = train_table_job_api.get(job_id).config.pquery_id
        assert pq_id is not None

        # NOTE the backend implementation currently handles sequentialization
        # between a training table future and a baseline job; that is, if the
        # training table future is still executing, the backend will wait on
        # the job ID completion before executing a baseline job. This preserves
        # semantics for both futures, ensures that Kumo works as expected if
        # used only via REST API, and allows us to avoid chaining calllbacks
        # in an ugly way here:
        api = global_state.client.baseline_job_api
        self._baseline_job_id = api.create(
            BaselineJobRequest(
                job_tags=dict(custom_tags),
                pquery_id=pq_id,
                metrics=self._metrics,
                graph_snapshot_id=graph.snapshot(non_blocking=non_blocking),
                train_table_job_id=job_id,
            ))
        out = BaselineJob(job_id=self._baseline_job_id)
        if non_blocking:
            return out
        return out.attach()



from __future__ import annotations

import asyncio
import logging
from concurrent.futures import Future
from datetime import datetime
from typing import List, Mapping, Optional, Union

import pandas as pd
from kumoapi.common import JobStatus
from kumoapi.jobs import GeneratePredictionTableJobResource, JobStatusReport
from typing_extensions import override

from kumoai import global_state
from kumoai.client.jobs import GeneratePredictionTableJobID
from kumoai.connector.s3_connector import S3URI
from kumoai.formatting import pretty_print_error_details
from kumoai.futures import KumoFuture, create_future

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL_S = 20


class PredictionTable:
    r"""A prediction table in the Kumo platform. A prediction table can
    either be initialized from a job ID of a completed prediction table
    generation job, or a path on a supported object store (S3 for a SaaS or
    Databricks deployment, and Snowflake session storage for Snowflake).

    .. warning::
        Custom prediction table is an experimental feature; please work
        with your Kumo POC to ensure you are using it correctly!


    .. code-block:: python

        import kumoai

        # Create a Prediction Table from a prediction table generation job.
        # Note that the job ID passed here must be in a completed state:
        prediction_table = kumoai.PredictionTable("gen-predtable-job-...")

        # Read the prediction table as a Pandas DataFrame:
        prediction_df = prediction_table.data_df()

        # Get URLs to download the prediction table:
        prediction_download_urls = prediction_table.data_urls()

    Args:
        job_id: ID of the prediction table generation job which
            generated this prediction table. If a custom table data path is
            specified, this parameter should be left as ``None``.
        table_data_path: S3 path of the table data location, for which Kumo
            must at least have read access. If a job ID is specified, this
            parameter should be left as ``None``.
    """
    def __init__(
        self,
        job_id: Optional[GeneratePredictionTableJobID] = None,
        table_data_path: Optional[str] = None,
    ) -> None:
        # Validation:
        if not (job_id or table_data_path):
            raise ValueError(
                "A PredictionTable must either be initialized with a table "
                "data path, or a job ID of a completed prediction table "
                "generation job.")
        if job_id and table_data_path:
            raise ValueError(
                "Please either pass a table data path, or a job ID of a "
                "completed prediction table generation job; passing both "
                "is not allowed.")

        # Custom path:
        self.table_data_uri: Optional[Union[str, S3URI]] = None
        if table_data_path is not None:
            if table_data_path.startswith('dbfs:/'):
                raise ValueError(
                    "Files from Databricks UC Volumes are not supported")
            if global_state.is_spcs:
                if table_data_path.startswith('s3://'):
                    raise ValueError(
                        "SPCS does not support S3 paths for prediction tables."
                    )
                # TODO(zeyuan): support custom stage path on SPCS:
                self.table_data_uri = table_data_path
            else:
                self.table_data_uri = S3URI(table_data_path).validate()

        # Job ID:
        self.job_id = job_id
        if job_id:
            status = _get_status(job_id).status
            if status != JobStatus.DONE:
                raise ValueError(
                    f"Job {job_id} is not yet complete (status: {status}). If "
                    f"you would like to create a future (waiting for "
                    f"prediction table generation success), please use "
                    f"`PredictionTableJob`.")

    def data_urls(self) -> List[str]:
        r"""Returns a list of URLs that can be used to view generated
        prediction table data; if a custom data path was passed, this path is
        simply returned.

        The list will contain more than one element if the table is
        partitioned; paths will be relative to the location of the Kumo data
        plane.
        """
        api = global_state.client.generate_prediction_table_job_api
        if not self.job_id:
            # Custom prediction table:
            if global_state.is_spcs:
                assert isinstance(self.table_data_uri, str)
                return [self.table_data_uri]
            else:
                assert isinstance(self.table_data_uri, S3URI)
                return [self.table_data_uri.uri]
        return api.get_table_data(self.job_id, presigned=True)

    def data_df(self) -> pd.DataFrame:
        r"""Returns a Pandas DataFrame object representing the generated
        or custom-specified prediction table data.

        .. warning::

            This method will load the full prediction table into memory as a
            :class:`~pandas.DataFrame` object. If you are working on a machine
            with limited resources, please use
            :meth:`~kumoai.pquery.PredictionTable.data_urls` instead to
            download the data and perform analysis per-partition.
        """
        if global_state.is_spcs:
            from kumoai.spcs import _parquet_dataset_to_df

            # TODO(dm): return type hint is wrong
            return _parquet_dataset_to_df(self.data_urls())
        else:
            urls = self.data_urls()
            try:
                return pd.concat([pd.read_parquet(x) for x in urls])
            except Exception as e:
                raise ValueError(
                    f"Could not create a Pandas DataFrame object from data "
                    f"paths {urls}. Please construct the DataFrame manually."
                ) from e

    @property
    def anchor_time(self) -> Optional[datetime]:
        r"""Returns the anchor time corresponding to the generated prediction
        table data, if the data was not custom-specified.
        """
        if self.job_id is None:
            logger.warning(
                "Fetching the anchor time is not supported for a custom "
                "prediction table (path: %s)", self.table_data_uri)
            return None
        api = global_state.client.generate_prediction_table_job_api
        return api.get_anchor_time(self.job_id)


# Prediction Table Future #####################################################


class PredictionTableJob(KumoFuture[PredictionTable]):
    r"""A representation of an ongoing prediction table generation job in the
    Kumo platform.

    .. code-block:: python

        import kumoai

        # See `PredictiveQuery` documentation:
        pquery = kumoai.PredictiveQuery(...)

        # If a prediction table is generated in nonblocking mode, the response
        # will be of type `PredictionTableJob`:
        prediction_table_job = pquery.generate_prediction_table(non_blocking=True)

        # You can also construct a `PredictionTableJob` from a job ID, e.g.
        # one that is present in the Kumo Jobs page:
        prediction_table_job = kumoai.PredictionTableJob("gen-predtable-job-...")

        # Get the status of the job:
        print(prediction_table_job.status())

        # Cancel the job:
        prediction_table_job.cancel()

        # Wait for the job to complete, and return a `PredictionTable`:
        prediction_table_job.result()

    Args:
        job_id: ID of the prediction table generation job.
    """  # noqa

    def __init__(
        self,
        job_id: GeneratePredictionTableJobID,
    ) -> None:
        self.job_id = job_id
        self.job: Optional[GeneratePredictionTableJobResource] = None
        # A training table holds a reference to the future that tracks the
        # execution of its generation.
        self._fut: Future = create_future(self._poll())

    @property
    def id(self) -> GeneratePredictionTableJobID:
        r"""The unique ID of this prediction table generation process."""
        return self.job_id

    @override
    def result(self) -> PredictionTable:
        return self._fut.result()

    @override
    def future(self) -> Future[PredictionTable]:
        return self._fut

    def status(self) -> JobStatusReport:
        r"""Returns the status of a running prediction table generation job."""
        return self._poll_job().job_status_report

    def cancel(self) -> None:
        r"""Cancels a running prediction table generation job, and raises an
        error if cancellation failed.
        """
        api = global_state.client.generate_prediction_table_job_api
        return api.cancel(self.job_id)

    def delete_tags(self, tags: List[str]) -> bool:
        r"""Removes the tags from the job.

        Args:
            tags (List[str]): The tags to remove.
        """
        api = (global_state.client.generate_prediction_table_job_api)
        return api.delete_tags(self.job_id, tags)

    def update_tags(self, tags: Mapping[str, Optional[str]]) -> bool:
        r"""Updates the tags of the job.

        Args:
            tags (Mapping[str, Optional[str]]): The tags to update.
                Note that the value 'none' will remove the tag. If the tag is
                not present, it will be added.
        """
        api = global_state.client.generate_prediction_table_job_api
        return api.update_tags(self.job_id, tags)

    # TODO(manan): make asynchronous natively with aiohttp:
    def _poll_job(self) -> GeneratePredictionTableJobResource:
        # Skip polling if job is already in terminal state.
        api = global_state.client.generate_prediction_table_job_api
        if not self.job or not self.job.job_status_report.status.is_terminal:
            self.job = api.get(self.job_id)
        return self.job

    async def _poll(self) -> PredictionTable:
        while not self.status().status.is_terminal:
            await asyncio.sleep(_DEFAULT_INTERVAL_S)
        api = global_state.client.generate_prediction_table_job_api
        status = self.status().status
        if status != JobStatus.DONE:
            error_details = api.get_job_error(self.job_id)
            error_str = pretty_print_error_details(error_details)
            raise RuntimeError(
                f"Prediction table generation for job {self.job_id} failed "
                f"with job status {status}. Encountered below"
                f" errors: {error_str}")
        return PredictionTable(self.job_id)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(job_id={self.job_id})'


def _get_status(job_id: str) -> JobStatusReport:
    api = global_state.client.generate_prediction_table_job_api
    resource: GeneratePredictionTableJobResource = api.get(job_id)
    return resource.job_status_report



from typing import TYPE_CHECKING, Any, Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from kumoai.client.endpoints import Endpoint, HTTPMethod

if TYPE_CHECKING:
    from kumoai.client.connector import ConnectorAPI
    from kumoai.client.graph import GraphAPI
    from kumoai.client.jobs import (
        ArtifactExportJobAPI,
        BaselineJobAPI,
        BatchPredictionJobAPI,
        GeneratePredictionTableJobAPI,
        GenerateTrainTableJobAPI,
        LLMJobAPI,
        TrainingJobAPI,
    )
    from kumoai.client.online import OnlineServingEndpointAPI
    from kumoai.client.pquery import PQueryAPI
    from kumoai.client.source_table import SourceTableAPI
    from kumoai.client.table import TableAPI

API_VERSION = 'v1'


class KumoClient:
    def __init__(
        self,
        url: str,
        api_key: Optional[str],
        spcs_token: Optional[str] = None,
    ) -> None:
        r"""Creates a client against the Kumo public API, provided a URL of
        the endpoint and an authentication token.

        Args:
            url: the public API endpoint URL.
            api_key: the public API authentication token.
            spcs_token: the SPCS token used for authentication to access the
                Kumo API endpoint.
        """
        self._url = url
        self._api_url = f"{url}/{API_VERSION}"
        self._api_key = api_key
        self._spcs_token = spcs_token

        retry_strategy = Retry(
            total=10,  # Maximum number of retries
            connect=3,  # How many connection-related errors to retry on
            read=3,  # How many times to retry on read errors
            status=5,  # How many times to retry on bad status codes (below)
            # Status codes to retry on.
            status_forcelist=[408, 429, 500, 502, 503, 504],
            # Exponential backoff factor: 2, 4, 8, 16 seconds delay)
            backoff_factor=2.0,
        )
        http_adapter = HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        session.mount('http://', http_adapter)
        session.mount('https://', http_adapter)
        self._session = session
        if self._api_key:
            self._session.headers.update({"X-API-Key": self._api_key})
        elif self._spcs_token:
            self._session.headers.update(
                {'Authorization': f'Snowflake Token={self._spcs_token}'})

    def authenticate(self) -> bool:
        r"""Raises an exception if authentication fails. Succeeds if the
        client is properly formed.
        """
        return self._session.get(f"{self._url}/v1/connectors").ok

    @property
    def artifact_export_api(self) -> 'ArtifactExportJobAPI':
        r"""Returns the artifact export API."""
        from kumoai.client.jobs import ArtifactExportJobAPI
        return ArtifactExportJobAPI(self)

    @property
    def connector_api(self) -> 'ConnectorAPI':
        r"""Returns the typed connector API."""
        from kumoai.client.connector import ConnectorAPI
        return ConnectorAPI(self)

    @property
    def source_table_api(self) -> 'SourceTableAPI':
        r"""Returns the typed source table API."""
        from kumoai.client.source_table import SourceTableAPI
        return SourceTableAPI(self)

    @property
    def table_api(self) -> 'TableAPI':
        r"""Returns the typed Kumo Table (snapshot) API."""
        from kumoai.client.table import TableAPI
        return TableAPI(self)

    @property
    def graph_api(self) -> 'GraphAPI':
        r"""Returns the typed Graph (metadata and snapshot) API."""
        from kumoai.client.graph import GraphAPI
        return GraphAPI(self)

    @property
    def pquery_api(self) -> 'PQueryAPI':
        r"""Returns the typed Predictive Query API."""
        from kumoai.client.pquery import PQueryAPI
        return PQueryAPI(self)

    @property
    def training_job_api(self) -> 'TrainingJobAPI':
        r"""Returns the typed Training Job API."""
        from kumoai.client.jobs import TrainingJobAPI
        return TrainingJobAPI(self)

    @property
    def batch_prediction_job_api(self) -> 'BatchPredictionJobAPI':
        from kumoai.client.jobs import BatchPredictionJobAPI
        return BatchPredictionJobAPI(self)

    @property
    def generate_train_table_job_api(self) -> 'GenerateTrainTableJobAPI':
        r"""Returns the typed Generate-Train-Table Job API."""
        from kumoai.client.jobs import GenerateTrainTableJobAPI
        return GenerateTrainTableJobAPI(self)

    @property
    def generate_prediction_table_job_api(
            self) -> 'GeneratePredictionTableJobAPI':
        from kumoai.client.jobs import GeneratePredictionTableJobAPI
        return GeneratePredictionTableJobAPI(self)

    @property
    def llm_job_api(self) -> 'LLMJobAPI':
        from kumoai.client.jobs import LLMJobAPI
        return LLMJobAPI(self)

    @property
    def baseline_job_api(self) -> 'BaselineJobAPI':
        r"""Returns the typed Training Job API."""
        from kumoai.client.jobs import BaselineJobAPI
        return BaselineJobAPI(self)

    @property
    def online_serving_endpoint_api(self) -> 'OnlineServingEndpointAPI':
        from kumoai.client.online import OnlineServingEndpointAPI
        return OnlineServingEndpointAPI(self)

    def _request(self, endpoint: Endpoint, **kwargs: Any) -> requests.Response:
        r"""Send a HTTP request to the specified endpoint."""
        endpoint_str = endpoint.get_path()
        if endpoint.method == HTTPMethod.GET:
            return self._get(endpoint_str, **kwargs)
        elif endpoint.method == HTTPMethod.POST:
            return self._post(endpoint_str, **kwargs)
        elif endpoint.method == HTTPMethod.PATCH:
            return self._patch(endpoint_str, **kwargs)
        elif endpoint.method == HTTPMethod.DELETE:
            return self._delete(endpoint_str, **kwargs)
        else:
            raise ValueError(f"Unsupported HTTP method: {endpoint.method}")

    def _get(self, endpoint: str, **kwargs: Any) -> requests.Response:
        r"""Send a GET request to the specified endpoint, with keyword
        arguments, returned objects, and exceptions raised corresponding to
        :meth:`requests.Session.get`.
        """
        url = self._format_endpoint_url(endpoint)
        return self._session.get(url=url, **kwargs)

    def _post(self, endpoint: str, **kwargs: Any) -> requests.Response:
        r"""Send a POST request to the specified endpoint, with keyword
        arguments, returned objects, and exceptions raised corresponding to
        :meth:`requests.Session.post`.
        """
        url = self._format_endpoint_url(endpoint)
        return self._session.post(url=url, **kwargs)

    def _patch(self, endpoint: str, **kwargs: Any) -> requests.Response:
        r"""Send a PATCH request to the specified endpoint, with keyword
        arguments, returned objects, and exceptions raised corresponding to
        :meth:`requests.Session.patch`.
        """
        url = self._format_endpoint_url(endpoint)
        return self._session.patch(url=url, **kwargs)

    def _delete(self, endpoint: str, **kwargs: Any) -> requests.Response:
        r"""Send a DELETE request to the specified endpoint, with keyword
        arguments, returned objects, and exceptions raised corresponding to
        :meth:`requests.Session.delete`.
        """
        url = self._format_endpoint_url(endpoint)
        return self._session.delete(url=url, **kwargs)

    def _format_endpoint_url(self, endpoint: str) -> str:
        if endpoint[0] == "/":
            endpoint = endpoint[1:]
        return f"{self._api_url}/{endpoint}"



import logging
import time
from typing import Any, Dict, Optional

import requests
from kumoapi.online_serving import (
    OnlinePredictionOptions,
    OnlineServingEndpointRequest,
    OnlineServingStatusCode,
)

from kumoai import global_state
from kumoai.experimental.rfm.local_graph import LocalGraph
from kumoai.trainer.online_serving import (
    OnlineServingEndpoint,
    OnlineServingEndpointFuture,
)

logger = logging.getLogger(__name__)

DEFAULT_GRAPH_NAME = "rfm_byoc_graph"


class KumoRFM:
    r"""The :class:`KumoRFM` class is an interface to the Kumo Relational
    Foundation model (RFM), see
    `KumoRFM <https://kumo.ai/research/kumo_relational_foundation_model.pdf>`_.

    KumoRFM is a relational foundation model, which can make generate
    predicitons in context for any relational dataset. The model is pretrained
    and the class provides an interface to query the model.

    The class is constructed from a :class:`LocalGraph` object.

    Example:
        .. code-block:: python

            # raw data
            df_users = pd.DataFrame(...)
            df_items = pd.DataFrame(...)
            df_transactions = pd.DataFrame(...)

            # construct LocalGraph from raw data
            graph = LocalGraph.from_data(
                {
                    'users': df_users,
                    'items': df_items,
                    'transactions': df_transactions,
                }
            )

            # start KumoRFM with this graph
            rfm = KumoRFM(graph)

            # query the model
            query_str = ("PREDICT COUNT(transactions.*, 0, 30, days) > 0 "
                         "FOR users.user_id=1")
            result = rfm.query(query_str)
            # Result is a pandas DataFrame with prediction probabilities
            print(result)  # user_id  COUNT(transactions.*, 0, 30, days) > 0
                           # 1        0.85

    """  # noqa: E501

    def __init__(self, graph: LocalGraph,
                 endpoint: Optional[OnlineServingEndpoint] = None) -> None:
        self.graph = graph
        self._endpoint: Optional[OnlineServingEndpoint] = endpoint
        self._endpoint_future: Optional[OnlineServingEndpointFuture] = None

        if self._endpoint is None:
            self._start()

    def query(
        self,
        query: str,
        wait_result_timeout: float = 10.0,
    ) -> Dict[str, Any]:
        """Query the RFM model with a query string

        Args:
            query: The RFM query string
            (e.g., "PREDICT COUNT(orders.*, 0, 30, days) > 0 "
            "FOR users.user_id=1")
            wait_result_timeout: Timeout in seconds to wait for the result

        Returns:
            Dictionary containing the prediction result
        """  # noqa: E501
        # Ensure endpoint is ready
        if not self._endpoint:
            try:
                if self._endpoint_future:
                    logger.info("Waiting for endpoint to be ready...")
                    self._endpoint = self._endpoint_future.result()
                else:
                    raise RuntimeError("Endpoint not initialized")
            except Exception as e:
                raise RuntimeError("Failed to get endpoint result") from e

        # Execute query via direct HTTP request to predict endpoint
        payload = {"query": query, 'wait_result_timeout': wait_result_timeout}

        resp = requests.post(self._endpoint._predict_url,
                             headers=global_state.client._session.headers,
                             json=payload)

        resp.raise_for_status()
        result = resp.json()

        return result

    def shutdown(self) -> None:
        r"""Clean up resources by destroying the RFM endpoint.

        This method attempts to destroy the RFM endpoint if one exists.

        Example:
            .. code-block:: python

                rfm = KumoRFM(graph)
                rfm.wait_until_ready()
                # ... use RFM ...
                rfm.shutdown()  # Clean up endpoint when done
        """  # noqa: E501
        if self._endpoint:
            logger.info("Found endpoint for KumoRFM, destroying...")
            try:
                self._endpoint.destroy()
            except Exception as e:
                logger.error(f"Failed to destroy endpoint: {e}")
        else:
            logger.info("No endpoint found, skipping...")

    def poll_status(self) -> Optional[OnlineServingStatusCode]:
        """Poll the current status of the RFM endpoint.

        Returns:
            The current status code of the endpoint, or None if no endpoint
            exists.

            Possible values:

            - OnlineServingStatusCode.IN_PROGRESS:
              Endpoint is being created/updated
            - OnlineServingStatusCode.READY: Endpoint is ready for queries
            - OnlineServingStatusCode.FAILED: Endpoint creation/update failed
        """
        if not self._endpoint_future:
            logger.info("Endpoint creation not started yet, "
                        "make sure to call _start() first")
            return None

        # Get the endpoint resource to check current status
        endpoint_api = global_state.client.online_serving_endpoint_api
        res = endpoint_api.get_if_exists(self._endpoint_future.id)

        if res is None:
            return None

        return res.status.status_code

    def is_ready(self) -> bool:
        """Check if the RFM endpoint is ready for queries.

        Returns:
            True if the endpoint is ready, False otherwise.
        """
        status = self.poll_status()
        return status == OnlineServingStatusCode.READY

    def wait_until_ready(self, timeout: Optional[float] = None,
                         sleep_interval: float = 10.0) -> None:
        """Wait until the RFM endpoint is ready.

        Args:
            timeout: Maximum time to wait in seconds. If None,
                waits indefinitely.
            sleep_interval: Time to sleep between polls in seconds.

        Raises:
            TimeoutError: If the endpoint is not ready within the
                timeout period.
            RuntimeError: If the endpoint failed to start.
        """
        start_time = time.time()

        if not self._endpoint_future:
            raise RuntimeError("Endpoint creation not started yet, "
                               "make sure to call _start() first")

        while True:
            endpoint_api = global_state.client.online_serving_endpoint_api
            res = endpoint_api.get_if_exists(self._endpoint_future.id)

            if res is None:
                raise RuntimeError("Endpoint resource not found")

            status = res.status.status_code

            if status == OnlineServingStatusCode.READY:
                logger.info(
                    "RFM endpoint is ready. Details:\n"
                    f"ID: {self._endpoint_future.id}\n"
                    f"URL: {res.endpoint_url}\n"
                    f"Launched at: {res.launched_at}\n"
                    f"Config: {res.config}\n"
                    f"Status: {res.status}\n"
                    f"Update status: {res.update if res.update else 'None'}")
                self._endpoint = self._endpoint_future.result()
                return
            elif status == OnlineServingStatusCode.FAILED:
                raise RuntimeError("RFM endpoint failed to start")
            elif timeout and (time.time() - start_time) > timeout:
                raise TimeoutError(
                    f"RFM endpoint not ready within {timeout} seconds")
            else:
                logger.info(
                    f"RFM endpoint not ready yet. Current status: {status}. ")

            time.sleep(sleep_interval)

    def _start(self) -> None:
        """Initialize KumoRFM by uploading data and creating endpoint"""
        self.graph.validate()
        kumo_graph = self.graph.to_kumo_graph()
        # skip validation as we validate the graph manually
        kumo_graph.save(DEFAULT_GRAPH_NAME, skip_validation=True)
        self._endpoint_future = self._create_endpoint()

    def _create_endpoint(self) -> OnlineServingEndpointFuture:
        """Create online serving endpoint for RFM"""
        endpoint_api = global_state.client.online_serving_endpoint_api

        request = OnlineServingEndpointRequest(
            model_training_job_id='FOUNDATION_MODEL',
            predict_options=OnlinePredictionOptions(),
        )

        # TODO(blaz): fix when we allow multiple graphs
        endpoint_id = endpoint_api.create(
            request,
            graph_name=DEFAULT_GRAPH_NAME,
            use_ge=False,
        )

        return OnlineServingEndpointFuture(endpoint_id)



import os
import sys
import threading
from dataclasses import dataclass
from typing import Any, Dict, Optional
import logging
import platform
from datetime import datetime

from kumoapi.typing import Dtype, Stype

from kumoai.client.client import KumoClient
from kumoai._logging import initialize_logging, _ENV_KUMO_LOG
from kumoai._singleton import Singleton
from kumoai.futures import create_future, initialize_event_loop
from kumoai.spcs import (
    _get_active_session,
    _get_spcs_token,
    _run_refresh_spcs_token,
)

initialize_logging()
initialize_event_loop()


@dataclass
class GlobalState(metaclass=Singleton):
    r"""Global storage of the state needed to create a Kumo client object. A
    singleton so its initialized state can be referenced elsewhere for free.
    """

    # NOTE fork semantics: CoW on Linux, and re-execed on Windows. So this will
    # likely not work on Windows unless we have special handling for the shared
    # state:
    _url: Optional[str] = None
    _api_key: Optional[str] = None
    _snowflake_credentials: Optional[Dict[str, Any]] = None
    _spcs_token: Optional[str] = None
    _snowpark_session: Optional[Any] = None

    thread_local: threading.local = threading.local()

    def clear(self) -> None:
        if hasattr(self.thread_local, '_client'):
            del self.thread_local._client
        self._url = None
        self._api_key = None
        self._snowflake_credentials = None
        self._spcs_token = None

    @property
    def initialized(self) -> bool:
        return self._url is not None and (
            self._api_key is not None or self._snowflake_credentials
            is not None or self._snowpark_session is not None)

    @property
    def client(self) -> KumoClient:
        r"""Accessor for the Kumo client. Note that clients are stored as
        thread-local variables as the requests Session library is not
        guaranteed to be thread-safe.

        For more information, see https://github.com/psf/requests/issues/1871.
        """
        if self._url is None or (self._api_key is None
                                 and self._spcs_token is None
                                 and self._snowpark_session is None):
            raise ValueError(
                "Client creation or authentication failed; please re-create "
                "your client before proceeding.")

        if hasattr(self.thread_local, '_client'):
            return self.thread_local._client

        client = KumoClient(self._url, self._api_key, self._spcs_token)
        self.thread_local._client = client
        return client

    @property
    def is_spcs(self) -> bool:
        return (self._api_key is None
                and (self._snowflake_credentials is not None
                     or self._snowpark_session is not None))


global_state: GlobalState = GlobalState()


def init(
    url: Optional[str] = None,
    api_key: Optional[str] = None,
    snowflake_credentials: Optional[Dict[str, str]] = None,
    snowflake_application: Optional[str] = None,
    log_level: str = "INFO",
) -> None:
    r"""Initializes and authenticates the API key against the Kumo service.
    Successful authentication is required to use the SDK.

    Example:
        >>> import kumoai
        >>> kumoai.init(url="<api_url>", api_key="<api_key>")  # doctest: +SKIP

    Args:
        url: The Kumo API endpoint. Can also be provided via the
            ``KUMO_API_ENDPOINT`` envronment variable. Will be inferred from
            the provided API key, if not provided.
        api_key: The Kumo API key. Can also be provided via the
            ``KUMO_API_KEY`` environment variable.
        snowflake_credentials: The Snowflake credentials to authenticate
            against the Kumo service. The dictionary should contain the keys
            ``"user"``, ``"password"``, and ``"account"``. This should only be
            provided for SPCS.
        snowflake_application: The Snowflake application.
        log_level: The logging level that Kumo operates under. Defaults to
            INFO; for more information, please see
            :class:`~kumoai.set_log_level`. Can also be set with the
            environment variable ``KUMOAI_LOG``.
    """  # noqa
    # Avoid mutations to the global state after it is set:
    if global_state.initialized:
        print(
            "Client has already been created. To re-initialize Kumo, please "
            "start a new interpreter. No changes will be made to the current "
            "session.")
        return

    set_log_level(os.getenv(_ENV_KUMO_LOG, log_level))

    # Get API key:
    api_key = api_key or os.getenv("KUMO_API_KEY")

    snowpark_session = None
    if snowflake_application:
        if url is not None:
            raise ValueError(
                "Client creation failed: both snowflake_application and url "
                "are specified. If running from a snowflake notebook, specify"
                "only snowflake_application.")
        snowpark_session = _get_active_session()
        if not snowpark_session:
            raise ValueError(
                "Client creation failed: snowflake_application is specified "
                "without an active snowpark session. If running outside "
                "a snowflake notebook, specify a URL and credentials.")
        description = snowpark_session.sql(
            f"DESCRIBE SERVICE {snowflake_application}."
            "USER_SCHEMA.KUMO_SERVICE").collect()[0]
        url = f"http://{description.dns_name}:8888/public_api"

    if api_key is None and not snowflake_application:
        if snowflake_credentials is None:
            raise ValueError(
                "Client creation failed: Neither API key nor snowflake "
                "credentials provided. Please either set the 'KUMO_API_KEY' "
                "or explicitly call `kumoai.init(...)`.")
        if (set(snowflake_credentials.keys())
                != {'user', 'password', 'account'}):
            raise ValueError(
                f"Provided credentials should be a dictionary with keys "
                f"'user', 'password', and 'account'. Only "
                f"{set(snowflake_credentials.keys())} were provided.")

    # Get or infer URL:
    url = url or os.getenv("KUMO_API_ENDPOINT")
    try:
        if api_key:
            url = url or f"http://{api_key.split(':')[0]}.kumoai.cloud/api"
    except KeyError:
        pass
    if url is None:
        raise ValueError(
            "Client creation failed: endpoint URL not provided. Please "
            "either set the 'KUMO_API_ENDPOINT' environment variable or "
            "explicitly call `kumoai.init(...)`.")

    # Assign global state after verification that client can be created and
    # authenticated successfully:
    spcs_token = _get_spcs_token(
        snowflake_credentials
    ) if not api_key and snowflake_credentials else None
    client = KumoClient(url=url, api_key=api_key, spcs_token=spcs_token)
    if client.authenticate():
        global_state._url = client._url
        global_state._api_key = client._api_key
        global_state._snowflake_credentials = snowflake_credentials
        global_state._spcs_token = client._spcs_token
        global_state._snowpark_session = snowpark_session
    else:
        raise ValueError("Client authentication failed. Please check if you "
                         "have a valid API key.")

    if not api_key and snowflake_credentials:
        # Refresh token every 10 minutes (expires in 1 hour):
        create_future(_run_refresh_spcs_token(minutes=10))

    logger = logging.getLogger('kumoai')
    log_level = logging.getLevelName(logger.getEffectiveLevel())
    logger.info(
        "Successfully initialized the Kumo SDK against deployment %s, with "
        "log level %s.", url, log_level)


def set_log_level(level: str) -> None:
    r"""Sets the Kumo logging level, which defines the amount of output that
    methods produce.

    Example:
        >>> import kumoai
        >>> kumoai.set_log_level("INFO")  # doctest: +SKIP

    Args:
        level: the logging level. Can be one of (in order of lowest to highest
            log output) :obj:`DEBUG`, :obj:`INFO`, :obj:`WARNING`,
            :obj:`ERROR`, :obj:`FATAL`, :obj:`CRITICAL`.
    """
    # logging library will ensure `level` is a valid string, and raise a
    # warning if not:
    logging.getLogger('kumoai').setLevel(level)


# Try to initialize purely with environment variables:
if ("pytest" not in sys.modules and "KUMO_API_KEY" in os.environ
        and "KUMO_API_ENDPOINT" in os.environ):
    init()

import kumoai.connector  # noqa
import kumoai.encoder  # noqa
import kumoai.graph  # noqa
import kumoai.pquery  # noqa
import kumoai.trainer  # noqa
import kumoai.utils  # noqa
import kumoai.databricks  # noqa

from kumoai.connector import (  # noqa
    SourceTable, SourceTableFuture, S3Connector, SnowflakeConnector,
    DatabricksConnector, BigQueryConnector, FileUploadConnector)
from kumoai.graph import Column, Edge, Graph, Table  # noqa
from kumoai.pquery import (  # noqa
    PredictionTableGenerationPlan, PredictiveQuery,
    TrainingTableGenerationPlan, TrainingTable, TrainingTableJob,
    PredictionTable, PredictionTableJob)
from kumoai.trainer import (  # noqa
    ModelPlan, Trainer, TrainingJobResult, TrainingJob,
    BatchPredictionJobResult, BatchPredictionJob)

__all__ = [
    'Dtype',
    'Stype',
    'SourceTable',
    'SourceTableFuture',
    'S3Connector',
    'SnowflakeConnector',
    'DatabricksConnector',
    'BigQueryConnector',
    'FileUploadConnector',
    'Column',
    'Table',
    'Graph',
    'Edge',
    'PredictiveQuery',
    'TrainingTable',
    'TrainingTableJob',
    'TrainingTableGenerationPlan',
    'PredictionTable',
    'PredictionTableJob',
    'PredictionTableGenerationPlan',
    'Trainer',
    'TrainingJobResult',
    'TrainingJob',
    'BatchPredictionJobResult',
    'BatchPredictionJob',
    'ModelPlan',
]

try:
    from importlib.metadata import version
    assert __package__ is not None
    __version__ = version(__package__)
except Exception:
    __version__ = '2.4'


def authenticate(api_url: str, redirect_port: int = 8765) -> str:
    """Starts a local HTTP server to handle OAuth2 or similar login flow, opens
    the browser for user login, and returns the access token.

    Args:
        api_url: The base URL for authentication (login page).
        redirect_port: The port for the local callback server (default: 8765).

    Returns:
        The access token as a string.
    """
    import http.server
    from socketserver import TCPServer
    import threading
    import webbrowser
    import urllib.parse
    import time
    from typing import Any, Dict

    logger = logging.getLogger('kumoai')

    token_status: Dict[str, Any] = {
        'token': None,
        'token_name': None,
        'failed': False
    }

    # Generate a token name
    token_name = (f"SDK_{platform.node()}_" +
                  datetime.now().strftime('%Y-%m-%d_%H:%M:%S'))

    class CallbackHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            parsed_path = urllib.parse.urlparse(self.path)
            params = urllib.parse.parse_qs(parsed_path.query)
            token = params.get('token', [None])[0]
            received_token_name = params.get('token_name', [None])[0]

            if token:
                token_status['token'] = token
                token_status['token_name'] = received_token_name
                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
            else:
                token_status['failed'] = True
                self.send_response(400)
                self.end_headers()

            html = '''
            <!DOCTYPE html>
            <html>
            <head>
                <title>Authenticate SDK</title>
                <style>
                    body {
                        margin: 0;
                        padding: 0;
                        display: flex;
                        justify-content: center;
                        align-items: center;
                        min-height: 100vh;
                        font-family:
                            -apple-system,
                            BlinkMacSystemFont,
                            'Segoe UI', Roboto, sans-serif;
                    }
                    .container {
                        text-align: center;
                        padding: 40px;
                    }
                    svg {
                        margin-bottom: 20px;
                    }
                    p {
                        font-size: 18px;
                        color: #333;
                    }
                </style>
            </head>
            <body>
                <div class="container">
                    <?xml version="1.0" encoding="UTF-8"?>
                    <svg xmlns="http://www.w3.org/2000/svg"
                        id="kumo-logo" width="183.908" height="91.586"
                        viewBox="0 0 183.908 91.586">
                        <g id="c">
                            <g id="Group_9893" data-name="Group 9893">
                                <path id="Path_4831" data-name="Path 4831"
                                    d="M67.159,67.919V46.238L53.494,59.491,
                                    68.862,82.3H61.567L49.1,63.74l-7.011,6.8V82.3h-6.02V29.605h6.02V62.182l16.642-16.36H73.109v22.1c0,5.453,3.611,9.419,9.277,9.419,5.547,0,9.14-3.9,9.2-9.282V0H0V91.586H91.586V80.317a15.7,15.7,0,0,1-9.2,2.828c-8.569,0-15.226-6.02-15.226-15.226Z"
                                    fill="#d40e8c">
                                </path>
                                <path id="Path_4832" data-name="Path 4832"
                                    d="M233.452,121.881h-6.019V98.3c0-4.745-3.117-8.286-7.932-8.286s-7.932,3.541-7.932,8.286v23.583h-6.02V98.3c0-4.745-3.116-8.286-7.932-8.286s-7.932,3.541-7.932,8.286v23.583h-6.02V98.51c0-7.932,5.736-14.023,13.952-14.023a12.106,12.106,0,0,1,10.906,6.02,12.3,12.3,0,0,1,10.978-6.02c8.285,0,13.951,6.091,13.951,14.023v23.37Z"
                                    transform="translate(-86.054 -39.585)"
                                    fill="#d40e8c">
                                </path>
                                <path id="Path_4833" data-name="Path 4833"
                                    d="M313.7,103.751c0,10.481-7.932,
                                    19.051-18.342,19.051-10.341,
                                    0-18.343-8.569-18.343-19.051,0-10.623,
                                    8-19.263,18.343-19.263C305.767,84.488,
                                    313.7,93.128,313.7,103.751Zm-6.02,
                                    0c0-7.436-5.523-13.527-12.322-13.527-6.728
                                    ,0-12.252,6.091-12.252,13.527,0,7.295,
                                    5.524,13.244,12.252,13.244,6.8,0,
                                    12.322-5.949,12.322-13.244Z"
                                    transform="translate(-129.791 -39.585)"
                                    fill="#d40e8c">
                                </path>
                            </g>
                        </g>
                    </svg>

                    <div id="success-div"
                        style="background: #f2f8f0;
                            border: 1px solid #1d8102;
                            border-radius: 1px;
                            padding: 24px 32px;
                            margin: 24px auto 0 auto;
                            max-width: 400px;
                            text-align: left;
                            display: none;"
                    >
                        <div style="font-size: 1.1em;
                            font-weight: bold;
                            margin-bottom: 10px;
                            text-align: left;"
                        >
                            Request successful
                        </div>
                        <div style="font-size: 1.1em;">
                            Kumo SDK has been granted a token.
                            You may now close this window.
                        </div>
                    </div>

                    <div id="failure-div"
                        style="background: #ffebeb;
                            border: 1px solid #ff837a;
                            border-radius: 1px;
                            padding: 24px 32px;
                            margin: 24px auto 0 auto;
                            max-width: 400px;
                            text-align: left;
                            display: none;"
                    >
                        <div style="font-size: 1.1em;
                            font-weight: bold;
                            margin-bottom: 10px;
                            text-align: left;"
                        >
                            Request failed
                        </div>
                        <div style="font-size: 1.1em;">
                            Failed to generate a token.
                            Please try again or contact Kumo
                            for further assistance.
                        </div>
                    </div>

                    <script>
                        // Show only the appropriate div based on the result
                        const search = window.location.search
                        const urlParams = new URLSearchParams(search);
                        const hasToken = urlParams.has('token');
                        if (hasToken) {
                            document
                                .getElementById('success-div')
                                .style.display = 'block';
                        } else {
                            document
                                .getElementById('failure-div')
                                .style.display = 'block';
                        }
                    </script>
                </div>
            </body>
            </html>
            '''
            self.wfile.write(html.encode('utf-8'))

        def log_message(self, format: str, *args: object) -> None:
            return  # Suppress logging

    # Find a free port if needed
    port = redirect_port
    for _ in range(10):
        try:
            with TCPServer(("", port), CallbackHandler) as _:
                break
        except OSError:
            port += 1
    else:
        raise RuntimeError(
            "Could not find a free port for the callback server.")

    # Start the server in a thread
    def serve() -> None:
        with TCPServer(("", port), CallbackHandler) as httpd:
            httpd.timeout = 60
            while token_status['token'] is None:
                httpd.handle_request()

    server_thread = threading.Thread(target=serve, daemon=True)
    server_thread.start()

    # Construct the login URL with callback_url and token_name
    callback_url = f"http://127.0.0.1:{port}/"
    login_url = (api_url +
                 f"?callback_url={urllib.parse.quote(callback_url)}" +
                 f"&token_name={urllib.parse.quote(token_name)}")
    webbrowser.open(login_url)

    # Wait for the token (timeout after 120s)
    start = time.time()
    while token_status['token'] is None and not token_status[
            'failed'] and time.time() - start < 120:
        time.sleep(1)

    if token_status['failed']:
        raise ValueError("Authentication failed.")

    if not isinstance(token_status['token'], str) or not token_status['token']:
        raise TimeoutError("Timed out waiting for authentication.")

    os.environ['KUMO_API_KEY'] = token_status['token']

    logger.info(f"Generated token \"{token_status['token_name']}\" " +
                "and saved to KUMO_API_KEY env variable")

    return token_status['token']



from .predictive_query import PredictiveQuery
from .training_table import (
    TrainingTable,
    TrainingTableJob,
)
from .prediction_table import (
    PredictionTable,
    PredictionTableJob,
)
from kumoapi.model_plan import (
    TrainingTableGenerationPlan,
    PredictionTableGenerationPlan,
    RunMode,
)

__all__ = [
    'RunMode',
    'PredictiveQuery',
    'TrainingTableGenerationPlan',
    'PredictionTableGenerationPlan',
    'TrainingTable',
    'TrainingTableJob',
    'PredictionTable',
    'PredictionTableJob',
]



from datetime import datetime
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from kumoapi.common import ValidationResponse
from kumoapi.jobs import (
    ArtifactExportRequest,
    ArtifactExportResponse,
    AutoTrainerProgress,
    BaselineJobRequest,
    BaselineJobResource,
    BatchPredictionJobResource,
    BatchPredictionRequest,
    CancelBatchPredictionJobResponse,
    CancelTrainingJobResponse,
    ErrorDetails,
    GeneratePredictionTableJobResource,
    GeneratePredictionTableRequest,
    GenerateTrainTableJobResource,
    GetEmbeddingsDfUrlResponse,
    GetPredictionsDfUrlResponse,
    JobRequestBase,
    JobResourceBase,
    JobStatus,
    PredictionProgress,
    TrainingJobRequest,
    TrainingJobResource,
    TrainingTableSpec,
)
from kumoapi.json_serde import from_json, to_json_dict
from kumoapi.source_table import LLMRequest, LLMResponse, SourceTableType
from kumoapi.train import TrainingStage
from typing_extensions import override

from kumoai.client import KumoClient
from kumoai.client.utils import (
    Returns,
    parse_patch_response,
    parse_response,
    raise_on_error,
)

TrainingJobID = str
BatchPredictionJobID = str
GenerateTrainTableJobID = str
GeneratePredictionTableJobID = str
LLMJobId = str
BaselineJobID = str

JobRequestType = TypeVar('JobRequestType', bound=JobRequestBase)
JobResourceType = TypeVar('JobResourceType', bound=JobResourceBase)


class CommonJobAPI(Generic[JobRequestType, JobResourceType]):
    def __init__(self, client: KumoClient, base_endpoint: str,
                 res_type: Type[JobResourceType]) -> None:
        self._client = client
        self._base_endpoint = base_endpoint
        self._res_type = res_type

    def create(self, request: JobRequestType) -> str:
        response = self._client._post(self._base_endpoint,
                                      json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(Dict[str, str], response)['id']

    def get(self, id: str) -> JobResourceType:
        response = self._client._get(f'{self._base_endpoint}/{id}')
        raise_on_error(response)
        return parse_response(self._res_type, response)

    def list(
        self,
        *,
        pquery_name: Optional[str] = None,
        pquery_id: Optional[str] = None,
        job_status: Optional[JobStatus] = None,
        limit: Optional[int] = None,
        additional_tags: Mapping[str, str] = {},
    ) -> List[JobResourceType]:
        params: Dict[str, Any] = {
            'pquery_name': pquery_name,
            'pquery_id': pquery_id,
            'job_status': job_status,
            'limit': limit
        }
        params.update(additional_tags)
        response = self._client._get(self._base_endpoint, params=params)
        raise_on_error(response)
        resource_elements = response.json()
        assert isinstance(resource_elements, list)
        return [from_json(e, self._res_type) for e in resource_elements]

    def delete_tags(self, job_id: str, tags: List[str]) -> bool:
        r"""Removes the tags from the job.

        Args:
            job_id (str): The ID of the job.
            tags (List[str]): The tags to remove.
        """
        return self.update_tags(job_id, {t: 'none' for t in tags})

    def update_tags(self, job_id: str,
                    custom_job_tags: Mapping[str, Optional[str]]) -> bool:
        r"""Updates the tags of the job.

        Args:
            job_id (str): The ID of the job.
            custom_job_tags (Mapping[str, Optional[str]]): The tags to update.
                Note that the value 'none' will remove the tag. If the tag is
                not present, it will be added.
        """
        response = self._client._patch(
            f'{self._base_endpoint}/{job_id}/tags',
            data=None,
            params={
                k: str(v)
                for k, v in custom_job_tags.items()
            },
        )
        raise_on_error(response)
        return parse_patch_response(response)


class BaselineJobAPI(CommonJobAPI[BaselineJobRequest, BaselineJobResource]):
    r"""Typed API definition for the baseline job resource."""
    def __init__(self, client: KumoClient) -> None:
        super().__init__(client, '/baseline_jobs', BaselineJobResource)


class TrainingJobAPI(CommonJobAPI[TrainingJobRequest, TrainingJobResource]):
    r"""Typed API definition for the training job resource."""
    def __init__(self, client: KumoClient) -> None:
        super().__init__(client, '/training_jobs', TrainingJobResource)

    def get_progress(self, id: TrainingJobID) -> AutoTrainerProgress:
        response = self._client._get(f'{self._base_endpoint}/{id}/progress')
        raise_on_error(response)
        return parse_response(AutoTrainerProgress, response)

    def holdout_data_url(self, id: TrainingJobID,
                         presigned: bool = True) -> str:
        response = self._client._get(f'{self._base_endpoint}/{id}/holdout',
                                     params={'presigned': presigned})
        raise_on_error(response)
        return response.text

    def cancel(self, id: str) -> CancelTrainingJobResponse:
        response = self._client._post(f'{self._base_endpoint}/{id}/cancel')
        raise_on_error(response)
        return parse_response(CancelTrainingJobResponse, response)


class BatchPredictionJobAPI(CommonJobAPI[BatchPredictionRequest,
                                         BatchPredictionJobResource]):
    r"""Typed API definition for the prediction job resource."""
    def __init__(self, client: KumoClient) -> None:
        super().__init__(client, '/prediction_jobs',
                         BatchPredictionJobResource)

    @override
    def create(self, request: BatchPredictionRequest) -> str:
        # TODO(manan): eventually, all `create` methods should
        # return a validation response:
        raise NotImplementedError

    def maybe_create(
        self, request: BatchPredictionRequest
    ) -> Tuple[Optional[str], ValidationResponse]:
        response = self._client._post(self._base_endpoint,
                                      json=to_json_dict(request))
        raise_on_error(response)
        return parse_response(
            Returns[Tuple[Optional[str], ValidationResponse]], response)

    def list(
        self,
        *,
        model_id: Optional[TrainingJobID] = None,
        pquery_name: Optional[str] = None,
        pquery_id: Optional[str] = None,
        job_status: Optional[JobStatus] = None,
        limit: Optional[int] = None,
        additional_tags: Mapping[str, str] = {},
    ) -> List[BatchPredictionJobResource]:
        if model_id:
            additional_tags = {**additional_tags, 'model_id': model_id}
        return super().list(pquery_name=pquery_name, pquery_id=pquery_id,
                            job_status=job_status, limit=limit,
                            additional_tags=additional_tags)

    def get_progress(self, id: str) -> PredictionProgress:
        response = self._client._get(f'{self._base_endpoint}/{id}/progress')
        raise_on_error(response)
        return parse_response(PredictionProgress, response)

    def cancel(self, id: str) -> CancelBatchPredictionJobResponse:
        response = self._client._post(f'{self._base_endpoint}/{id}/cancel')
        raise_on_error(response)
        return parse_response(CancelBatchPredictionJobResponse, response)

    def get_batch_predictions_url(self, id: str) -> List[str]:
        """Returns presigned URLs pointing to the locations where the
        predictions are stored. Depending on the environment where this is run,
        they could be AWS S3 paths, Snowflake stage paths, or Databricks UC
        volume paths.

        Args:
            id (str): ID of the batch prediction job for which predictions are
                requested
        """
        response = self._client._get(
            f'{self._base_endpoint}/{id}/get_prediction_df_urls')
        raise_on_error(response)
        return parse_response(
            GetPredictionsDfUrlResponse,
            response,
        ).prediction_partitions

    def get_batch_embeddings_url(self, id: str) -> List[str]:
        """Returns presigned URLs pointing to the locations where the
        embeddings are stored. Depending on the environment where this is run,
        they could be AWS S3 paths, Snowflake stage paths, or Databricks UC
        volume paths.

        Args:
            id (str): ID of the batch prediction job for which embeddings are
                requested
        """
        response = self._client._get(
            f'{self._base_endpoint}/{id}/get_embedding_df_urls')
        raise_on_error(response)
        return parse_response(
            GetEmbeddingsDfUrlResponse,
            response,
        ).embedding_partitions


class GenerateTrainTableJobAPI(CommonJobAPI[GeneratePredictionTableRequest,
                                            GenerateTrainTableJobResource]):
    r"""Typed API definition for training table generation job resource."""
    def __init__(self, client: KumoClient) -> None:
        super().__init__(client, '/gentraintable_jobs',
                         GenerateTrainTableJobResource)

    def get_table_data(self, id: GenerateTrainTableJobID,
                       presigned: bool = True) -> List[str]:
        """Return a list of URLs to access train table parquet data.
        There might be multiple URLs if the table data is partitioned into
        multiple files.
        """
        return self._get_table_data(id, presigned)

    def _get_table_data(self, id: GenerateTrainTableJobID,
                        presigned: bool = True,
                        raw_path: bool = False) -> List[str]:
        """Helper function to get train table data."""
        # Raw path to get local file path instead of SPCS stage path
        params: Dict[str, Any] = {'presigned': presigned, 'raw_path': raw_path}
        resp = self._client._get(f'{self._base_endpoint}/{id}/table_data',
                                 params=params)
        raise_on_error(resp)
        return parse_response(List[str], resp)

    def get_split_masks(
            self, id: GenerateTrainTableJobID) -> Dict[TrainingStage, str]:
        """Return a dictionary of presigned URLs keyed by training stage.
        Each URL points to a torch-serialized (default pickle protocol) file of
        the mask tensor for that training stage.

        Example:
            >>> # code to load a mask tensor:
            >>> import io
            >>> import torch
            >>> import requests
            >>> masks = get_split_masks('some-gen-traintable-job-id')
            >>> data_bytes = requests.get(masks[TrainingStage.TEST]).content
            >>> test_mask_tensor = torch.load(io.BytesIO(data))
        """
        resp = self._client._get(f'{self._base_endpoint}/{id}/split_masks')
        raise_on_error(resp)
        return parse_response(Dict[TrainingStage, str], resp)

    def get_progress(self, id: str) -> Dict[str, int]:
        response = self._client._get(f'{self._base_endpoint}/{id}/progress')
        raise_on_error(response)
        return parse_response(Dict[str, int], response)

    def cancel(self, id: str) -> None:
        response = self._client._post(f'{self._base_endpoint}/{id}/cancel')
        raise_on_error(response)

    def validate_custom_train_table(
        self,
        id: str,
        source_table_type: SourceTableType,
        train_table_mod: TrainingTableSpec,
    ) -> ValidationResponse:
        response = self._client._post(
            f'{self._base_endpoint}/{id}/validate_custom_train_table',
            json=to_json_dict({
                'custom_table': source_table_type,
                'train_table_mod': train_table_mod,
            }),
        )
        return parse_response(ValidationResponse, response)

    def get_job_error(self, id: str) -> ErrorDetails:
        """Thin API wrapper for fetching errors from the jobs.

        Arguments:
        id (str): Id of the job whose related errors are expected to be
            queried.
        """
        response = self._client._get(f'{self._base_endpoint}/{id}/get_errors')
        raise_on_error(response)
        return parse_response(ErrorDetails, response)


class GeneratePredictionTableJobAPI(
        CommonJobAPI[GeneratePredictionTableRequest,
                     GeneratePredictionTableJobResource]):
    r"""Typed API definition for prediction table generation job resource."""
    def __init__(self, client: KumoClient) -> None:
        super().__init__(client, '/genpredtable_jobs',
                         GeneratePredictionTableJobResource)

    def get_anchor_time(self, id: BatchPredictionJobID) -> Optional[datetime]:
        response = self._client._get(
            f'{self._base_endpoint}/{id}/get_anchor_time')
        raise_on_error(response)
        return parse_response(Returns[Optional[datetime]], response)

    def get_table_data(self, id: GeneratePredictionTableJobID,
                       presigned: bool = True) -> List[str]:
        """Return a list of URLs to access prediction table parquet data.
        There might be multiple URLs if the table data is partitioned into
        multiple files.
        """
        params: Dict[str, Any] = {'presigned': presigned}
        resp = self._client._get(f'{self._base_endpoint}/{id}/table_data',
                                 params=params)
        raise_on_error(resp)
        return parse_response(List[str], resp)

    def cancel(self, id: str) -> None:
        response = self._client._post(f'{self._base_endpoint}/{id}/cancel')
        raise_on_error(response)

    def get_job_error(self, id: str) -> ErrorDetails:
        """Thin API wrapper for fetching errors from the jobs.

        Arguments:
        id (str): Id of the job whose related errors are expected to be
            queried.
        """
        response = self._client._get(f'{self._base_endpoint}/{id}/get_errors')
        raise_on_error(response)
        return parse_response(ErrorDetails, response)


class LLMJobAPI:
    r"""Typed API definition for LLM job resource."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client
        self._base_endpoint = '/llm_embedding_job'

    def create(self, request: LLMRequest) -> LLMJobId:
        response = self._client._post(
            self._base_endpoint,
            json=to_json_dict(request),
        )
        raise_on_error(response)
        return parse_response(LLMResponse, response).job_id

    def get(self, id: LLMJobId) -> JobStatus:
        response = self._client._get(f'{self._base_endpoint}/status/{id}')
        raise_on_error(response)
        return parse_response(JobStatus, response)

    def cancel(self, id: LLMJobId) -> JobStatus:
        response = self._client._delete(f'{self._base_endpoint}/cancel/{id}')
        return response


class ArtifactExportJobAPI:
    r"""Typed API definition for artifact export job resource."""
    def __init__(self, client: KumoClient) -> None:
        self._client = client
        self._base_endpoint = '/artifact'

    def create(self, request: ArtifactExportRequest) -> str:
        response = self._client._post(
            self._base_endpoint,
            json=to_json_dict(request),
        )
        raise_on_error(response)
        return parse_response(ArtifactExportResponse, response).job_id

    # TODO Add an API in artifact export to get
    # JobStatusReport and not just JobStatus
    def get(self, id: str) -> JobStatus:
        response = self._client._get(f'{self._base_endpoint}/{id}')
        raise_on_error(response)
        return parse_response(JobStatus, response)

    def cancel(self, id: str) -> JobStatus:
        response = self._client._post(f'{self._base_endpoint}/{id}/cancel')
        raise_on_error(response)
        return parse_response(JobStatus, response)



import os
from typing import List, Optional, Set, Tuple, Union

from kumoapi.jobs import (
    BigQueryPredictionOutput,
    DatabricksPredictionOutput,
    MetadataField,
    PredictionArtifactType,
    PredictionOutputConfig,
    PredictionStorageType,
    S3PredictionOutput,
    SnowflakePredictionOutput,
    WriteMode,
)

from kumoai.artifact_export.config import OutputConfig
from kumoai.connector import (
    BigQueryConnector,
    Connector,
    DatabricksConnector,
    S3Connector,
    SnowflakeConnector,
)
from kumoai.databricks import DB_SEP


def validate_output_arguments(
    output_types: Set[str],
    output_connector: Optional[Connector] = None,
    output_table_name: Optional[str] = None,
) -> None:
    r"""Validate the output arguments for a prediction job or an export job."""
    output_types = {x.lower() for x in output_types}
    assert output_types.issubset({'predictions', 'embeddings'})
    if output_connector is not None:
        assert output_table_name is not None
        assert isinstance(output_connector,
                          (S3Connector, SnowflakeConnector,
                           DatabricksConnector, BigQueryConnector))
        if not isinstance(output_table_name, str):
            raise ValueError(
                f"The output table name must be a string for all "
                f"non-Databricks connectors. Got '{output_table_name}'.")

        if isinstance(output_connector, S3Connector):
            assert output_connector.root_dir is not None
        if isinstance(output_connector, DatabricksConnector):
            assert DB_SEP in output_table_name


def build_prediction_output_config(
    output_type: str,
    output_connector: Optional[Connector] = None,
    output_table_name: Optional[Union[str, Tuple]] = None,
    output_metadata_fields: Optional[List[MetadataField]] = None,
    output_config: Optional[OutputConfig] = None,
) -> PredictionOutputConfig:
    r"""Build the prediction output config."""
    assert output_config is not None
    artifact_type = PredictionArtifactType(output_type.upper())
    output_name = f"{output_table_name}_{output_type}"
    output_metadata_fields = output_metadata_fields or []
    if isinstance(output_connector, S3Connector):
        assert output_connector.root_dir is not None
        return S3PredictionOutput(
            artifact_type=artifact_type,
            file_path=os.path.join(output_connector.root_dir, output_name),
            extra_fields=output_metadata_fields,
        )
    elif isinstance(output_connector, SnowflakeConnector):
        return SnowflakePredictionOutput(
            artifact_type=artifact_type,
            connector_id=output_connector.name,
            table_name=output_name,
            extra_fields=output_metadata_fields,
            write_mode=output_config.connector_specific_config.write_mode
            if output_config.connector_specific_config is not None else
            WriteMode.OVERWRITE,
        )
    elif isinstance(output_connector, DatabricksConnector):
        return DatabricksPredictionOutput(
            artifact_type=artifact_type,
            connector_id=output_connector.name,
            table_name=output_name,
            extra_fields=output_metadata_fields,
        )
    elif isinstance(output_connector, BigQueryConnector):
        return BigQueryPredictionOutput(
            storage_type=PredictionStorageType.BIGQUERY,
            artifact_type=artifact_type,
            connector_id=output_connector.name,
            table_name=output_name,
            extra_fields=output_metadata_fields,
            write_mode=output_config.connector_specific_config.write_mode
            if output_config.connector_specific_config is not None else
            WriteMode.OVERWRITE,
        )
    else:
        raise NotImplementedError()



from typing import List

from kumoapi.source_table import (
    DataSourceType,
    FileType,
    S3SourceTableRequest,
    SourceTableConfigRequest,
    SourceTableConfigResponse,
)
from typing_extensions import override

from kumoai import global_state
from kumoai.connector.base import Connector


class FileUploadConnector(Connector):
    r"""Defines a connector to files directly uploaded to Kumo, either as
    'parquet' or 'csv' (non-partitioned) data.

    To get started with file upload, please first upload a table with
    the :meth:`~kumoai.connector.upload_table` method. You can then access
    this table behind the file upload connector as follows:

    .. code-block:: python

        import kumoai
        from kumoai.connector import upload_table

        # Upload the table; assume it is stored at `/data/users.parquet`
        upload_table(name="users", path="/data/users.parquet")

        # Create the file upload connector:
        connector = kumoai.FileUploadConnector(file_type="parquet")

        # Check that the file upload connector has a `users` table:
        assert connector.has_table("users")

    Args:
        file_type: The file type of uploaded data. Can be either ``"csv"``
            or ``"parquet"``.
    """
    def __init__(self, file_type: str) -> None:
        r"""Creates the connector to uploaded files of type
        :obj:`file_type`.
        """
        assert file_type.lower() in {'parquet', 'csv'}
        self._file_type = file_type.lower()

    @property
    def name(self) -> str:
        return f'{self._file_type}_upload_connector'

    @override
    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.S3

    @property
    def file_type(self) -> FileType:
        return (FileType.PARQUET
                if self._file_type == 'parquet' else FileType.CSV)

    def _get_table_config(self, table_name: str) -> SourceTableConfigResponse:
        req = SourceTableConfigRequest(connector_id=self.name,
                                       table_name=table_name,
                                       source_type=self.source_type,
                                       file_type=None)
        return global_state.client.source_table_api.get_table_config(req)

    @override
    def _source_table_request(self,
                              table_names: List[str]) -> S3SourceTableRequest:
        return S3SourceTableRequest(s3_root_dir="", connector_id=self.name,
                                    table_names=table_names, file_type=None)



from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
import pyarrow as pa
from kumoapi.typing import Dtype, Stype
from pandas.api.types import (
    is_datetime64_any_dtype,
    is_list_like,
    is_object_dtype,
)

from kumoai.experimental.rfm.infer import (
    contains_categorical,
    contains_id,
    contains_multicategorical,
    contains_timestamp,
)

# Maximum number of rows to check for dtype inference in object columns
_MAX_NUM_ROWS_FOR_DTYPE_INFERENCE = 100

# Mapping from pandas/numpy dtypes to Kumo Dtypes
PANDAS_TO_DTYPE: Dict[Any, Dtype] = {
    np.dtype('bool'): Dtype.bool,
    pd.BooleanDtype(): Dtype.bool,
    pa.bool_(): Dtype.bool,
    np.dtype('byte'): Dtype.int32,
    pd.UInt8Dtype(): Dtype.int32,
    np.dtype('int16'): Dtype.int32,
    pd.Int16Dtype(): Dtype.int32,
    np.dtype('int32'): Dtype.int32,
    pd.Int32Dtype(): Dtype.int32,
    np.dtype('int64'): Dtype.int64,
    pd.Int64Dtype(): Dtype.int64,
    np.dtype('float32'): Dtype.float32,
    pd.Float32Dtype(): Dtype.float32,
    np.dtype('float64'): Dtype.float64,
    pd.Float64Dtype(): Dtype.float64,
    np.dtype('object'): Dtype.string,
    pd.StringDtype(storage='python'): Dtype.string,
    pd.StringDtype(storage='pyarrow'): Dtype.string,
    pa.string(): Dtype.string,
    pa.binary(): Dtype.binary,
    np.dtype('datetime64[ns]'): Dtype.date,
    np.dtype('timedelta64[ns]'): Dtype.timedelta,
    pa.list_(pa.float32()): Dtype.floatlist,
    pa.list_(pa.int64()): Dtype.intlist,
    pa.list_(pa.string()): Dtype.stringlist,
}

# Mapping from Kumo Dtype to dtype families
DTYPE_TO_FAMILY: Dict[Dtype, str] = {
    # Number family - includes booleans, integers, floats, and numeric lists
    Dtype.byte:
    "number",
    Dtype.int:
    "number",
    Dtype.int16:
    "number",
    Dtype.int32:
    "number",
    Dtype.int64:
    "number",
    Dtype.float:
    "number",
    Dtype.float32:
    "number",
    Dtype.float64:
    "number",

    # String family - includes text and string lists
    Dtype.string:
    "string",
    Dtype.stringlist:
    "string",

    # Other family - includes binary, dates, and time-related types
    Dtype.bool:
    "other",
    Dtype.floatlist:
    "other",
    Dtype.intlist:
    "other",
    Dtype.binary:
    "other",
    Dtype.date:
    "other",
    Dtype.time:
    "other",
    Dtype.timedelta:
    "other",
    Dtype.unsupported:
    "other",
}


def dtype_to_family(dtype: Dtype) -> str:
    """Convert a Kumo Dtype to its dtype family.

    Args:
        dtype: The Kumo Dtype to convert

    Returns:
        The dtype family as a string: "number", "string", or "other"

    Raises:
        KeyError: If the dtype is not found in the mapping
    """
    return DTYPE_TO_FAMILY[dtype]


def to_dtype(dtype: Any, ser: Optional[pd.Series] = None) -> Dtype:
    """Convert a pandas/numpy/pyarrow dtype to Kumo Dtype.

    Args:
        dtype: The dtype to convert
        ser: Optional pandas Series for additional type inference

    Returns:
        The corresponding Kumo Dtype
    """
    if is_datetime64_any_dtype(dtype):
        return Dtype.date
    if isinstance(dtype, pd.CategoricalDtype):
        return Dtype.string
    if is_object_dtype(dtype) and ser is not None and len(ser) > 0:
        if is_list_like(ser.iloc[0]):
            # the 0-th element might be an empty list or a list of N/A so
            # iterate over the first _MAX_NUM_ROWS_FOR_DTYPE_INFERENCE elements
            # to infer the dtype:
            for i in range(min(len(ser), _MAX_NUM_ROWS_FOR_DTYPE_INFERENCE)):
                # pd.isna can't be called on a list
                if (not isinstance(ser.iloc[i], list)
                        and not isinstance(ser.iloc[i], np.ndarray)
                        and pd.isna(ser.iloc[i])) or len(ser.iloc[i]) == 0:
                    continue
                elif isinstance(ser.iloc[i][0], float):
                    return Dtype.floatlist
                elif np.issubdtype(type(ser.iloc[i][0]), int):
                    return Dtype.intlist
                elif isinstance(ser.iloc[i][0], str):
                    return Dtype.stringlist
    return PANDAS_TO_DTYPE[dtype]


# TODO(blaz): refine and test
# Mapping from Kumo Dtype to Kumo Stype
DTYPE_TO_STYPE: Dict[Dtype, Stype] = {
    # Boolean types - always categorical
    Dtype.bool:
    Stype.categorical,

    # Integer types - numerical by default, but can be ID or categorical
    # based on context
    Dtype.int:
    Stype.numerical,
    Dtype.byte:
    Stype.numerical,
    Dtype.int16:
    Stype.numerical,
    Dtype.int32:
    Stype.numerical,
    Dtype.int64:
    Stype.numerical,

    # Float types - always numerical
    Dtype.float:
    Stype.numerical,
    Dtype.float32:
    Stype.numerical,
    Dtype.float64:
    Stype.numerical,

    # String types - text by default, but can be categorical based on context
    Dtype.string:
    Stype.text,

    # Datetime types - always timestamp
    Dtype.date:
    Stype.timestamp,
    Dtype.time:
    Stype.timestamp,
    Dtype.timedelta:
    Stype.timestamp,

    # List types - sequence
    Dtype.floatlist:
    Stype.sequence,
    Dtype.intlist:
    Stype.sequence,
    Dtype.stringlist:
    Stype.sequence,

    # Binary and unsupported types
    Dtype.binary:
    Stype.unsupported,
    Dtype.unsupported:
    Stype.unsupported,
}


# TODO(blaz): refine and test
def infer_stype(ser: pd.Series, column_name: str, dtype: Dtype) -> Stype:
    r"""Infers the semantic type of a column.

    Args:
        ser: A :class:`pandas.Series` to analyze.
        column_name: The name of the column (used for pattern matching)
        dtype: The data type.

    Returns:
        The semantic type.
    """
    if contains_id(ser, column_name, dtype):
        return Stype.ID

    if contains_timestamp(ser, column_name, dtype):
        return Stype.timestamp

    if contains_multicategorical(ser, column_name, dtype):
        return Stype.multicategorical

    if contains_categorical(ser, column_name, dtype):
        return Stype.categorical

    return dtype.default_stype


# TODO(blaz): refine and test
def detect_primary_key(df: pd.DataFrame) -> Optional[str]:
    """Auto-detect potential primary key column.

    Args:
        df: The pandas DataFrame to analyze

    Returns:
        The name of the detected primary key column, or None if not found
    """
    for col in df.columns:
        col_lower = col.lower()

        # Check naming patterns
        if any(pattern in col_lower for pattern in ['id', 'key', 'primary']):
            # Check if values are unique and non-null
            if df[col].nunique() == len(df) and df[col].notna().all():
                return col

    return None



from kumoapi.jobs import ErrorDetails


def pretty_print_error_details(error_details: ErrorDetails) -> str:
    """Pretty prints the ErrorDetails combining all the individual items.
    If there are CTAs, they are also displayed after creating
    corresponding hyperlinks.

    Arguments:
    error_details (ErrorDetails): Standard ErrorDetails response from
        get_errors APIs.
    """
    out = ""
    ctr = None
    if len(error_details.items) != 1:
        out += "Encountered multiple errors:\n"
        ctr = 1
    for error_detail in error_details.items:
        if ctr is not None:
            out += f'{ctr}.'
            ctr += 1
        if error_detail.title is not None:
            out += f'{error_detail.title}: '
        out += error_detail.description
        if error_detail.cta is not None:
            out += 'Follow the link for potential resolution:'
            f' {error_detail.cta.url}'
        out += '\n'

    return out



from setuptools import setup

setup(
    # Dynamic specification in pyproject.toml:
    install_requires=[
        "pandas",
        "requests>=2.28.2",
        "urllib3",
        "plotly",
        "typing_extensions>=4.5.0",  # for override
        "kumo-api==0.16.0",
        "tqdm>=4.66.0",
        "aiohttp>=3.10.0",
        "pyarrow>=8.0.0",
    ])



from __future__ import annotations

import asyncio
import logging
import os
import time
from concurrent.futures import Future
from typing import List, Mapping, Optional, Tuple, Union

import pandas as pd
from kumoapi.common import JobStatus
from kumoapi.jobs import (
    ArtifactExportRequest,
    CustomTrainingTable,
    GenerateTrainTableJobResource,
    JobStatusReport,
    SourceTableType,
    TrainingTableOutputConfig,
    TrainingTableSpec,
    WriteMode,
)
from kumoapi.source_table import S3SourceTable
from tqdm.auto import tqdm
from typing_extensions import Self, override

from kumoai import global_state
from kumoai.artifact_export import (
    ArtifactExportJob,
    ArtifactExportResult,
    TrainingTableExportConfig,
)
from kumoai.client.jobs import (
    GenerateTrainTableJobAPI,
    GenerateTrainTableJobID,
)
from kumoai.connector import S3Connector, SourceTable
from kumoai.databricks import to_db_table_name
from kumoai.formatting import pretty_print_error_details
from kumoai.futures import KumoProgressFuture, create_future

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL_S = 20


class TrainingTable:
    r"""A training table in the Kumo platform. A training table can be
    initialized from a job ID of a completed training table generation job.

    .. code-block:: python

        import kumoai

        # Create a Training Table from a training table generation job. Note
        # that the job ID passed here must be in a completed state:
        training_table = kumoai.TrainingTable("gen-traintable-job-...")

        # Read the training table as a Pandas DataFrame:
        training_df = training_table.data_df()

        # Get URLs to download the training table:
        training_download_urls = training_table.data_urls()

        # Add weight column to the training table:
        # see `kumo-sdk.examples.datasets.weighted_train_table.py`
        # for a more detailed example
        # 1. Export train table
        connector = kumo.S3Connector("s3_path")
        training_table.export(TrainingTableExportConfig(
            output_types={'training_table'},
            output_connector=connector,
            output_table_name="<any_name>"))
        # 2. Assume the weight column was added to the train table
        # and it was saved to the same S3 path as "<mod_name>"
        training_table.update(SourceTable("<mod_table>", connector),
                              TrainingTableSpec(weight_col="weight"))

    Args:
        job_id: ID of the training table generation job which generated this
            training table.
    """
    def __init__(self, job_id: GenerateTrainTableJobID):
        self.job_id = job_id
        status = _get_status(job_id).status
        self._custom_train_table: Optional[CustomTrainingTable] = None
        if status != JobStatus.DONE:
            raise ValueError(
                f"Job {job_id} is not yet complete (status: {status}). If you "
                f"would like to create a future (waiting for training table "
                f"generation success), please use `TrainingTableJob`.")

    def data_urls(self) -> List[str]:
        r"""Returns a list of URLs that can be used to view generated
        training table data. The list will contain more than one element
        if the table is partitioned; paths will be relative to the location of
        the Kumo data plane.
        """
        api: GenerateTrainTableJobAPI = (
            global_state.client.generate_train_table_job_api)
        return api._get_table_data(self.job_id, presigned=True, raw_path=True)

    def data_df(self) -> pd.DataFrame:
        r"""Returns a :class:`~pandas.DataFrame` object representing the
        generated training data.

        .. warning::

            This method will load the full training table into memory as a
            :class:`~pandas.DataFrame` object. If you are working on a machine
            with limited resources, please use
            :meth:`~kumoai.pquery.TrainingTable.data_urls` instead to download
            the data and perform analysis per-partition.
        """
        urls = self.data_urls()
        if global_state.is_spcs:
            from kumoai.spcs import _parquet_dataset_to_df

            # TODO(dm): return type hint is wrong
            return _parquet_dataset_to_df(self.data_urls())

        try:
            return pd.concat([pd.read_parquet(x) for x in urls])
        except Exception as e:
            raise ValueError(
                f"Could not create a Pandas DataFrame object from data paths "
                f"{urls}. Please construct the DataFrame manually.") from e

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(job_id={self.job_id})'

    def _to_s3_api_source_table(self,
                                source_table: SourceTable) -> S3SourceTable:
        assert isinstance(source_table.connector, S3Connector)
        source_type = source_table._to_api_source_table()
        root_dir: str = source_table.connector.root_dir  # type: ignore
        if root_dir[-1] != os.sep:
            root_dir = root_dir + os.sep
        return S3SourceTable(
            s3_path=root_dir,
            source_table_name=source_table.name,
            file_type=source_type.file_type,
        )

    def export(
        self,
        output_config: TrainingTableExportConfig,
        non_blocking: bool = True,
    ) -> Union[ArtifactExportJob, ArtifactExportResult]:
        r"""Export the training table to the connector.
        specified in the output config. Use the exported table to
        add a weight column then use `update` to update the training table.

        Args:
            output_config: The output configuration to write the training
                table.
            non_blocking: If ``True``, the method will return a future object
                `ArtifactExportJob` representing the export job.
                If ``False``, the method will block until the export job is
                complete and return `ArtifactExportResult`.
        """
        assert output_config.output_connector is not None
        assert output_config.output_types == {'training_table'}
        output_table_name = to_db_table_name(output_config.output_table_name)
        assert output_table_name is not None
        s3_path = None
        connector_id = None
        table_name = ""
        write_mode = WriteMode.OVERWRITE

        if isinstance(output_config.output_connector, S3Connector):
            assert output_config.output_connector.root_dir is not None
            s3_path = output_config.output_connector.root_dir
            s3_path = os.path.join(s3_path, output_table_name)
        else:
            connector_id = output_config.output_connector.name
            table_name = output_table_name
            if output_config.connector_specific_config:
                write_mode = output_config.connector_specific_config.write_mode

        api = global_state.client.artifact_export_api
        output_config = TrainingTableOutputConfig(
            s3_path=s3_path,
            connector_id=connector_id,
            table_name=table_name,
            write_mode=write_mode,
        )

        request = ArtifactExportRequest(job_id=self.job_id,
                                        training_table_output=output_config)
        job_id = api.create(request)
        if non_blocking:
            return ArtifactExportJob(job_id)
        return ArtifactExportJob(job_id).attach()

    def validate_custom_table(self, source_table_type: SourceTableType,
                              train_table_mod: TrainingTableSpec) -> None:
        r"""Validates the modified training table.

        Args:
            source_table (SourceTable): The source table to be used as the
                modified training table.
            train_table_mod (TrainTableSpec): The modification specification.

        Raises:
            ValueError: If the modified training table is invalid.

        """
        api: GenerateTrainTableJobAPI = (
            global_state.client.generate_train_table_job_api)
        response = api.validate_custom_train_table(self.job_id,
                                                   source_table_type,
                                                   train_table_mod)
        if not response.ok:
            raise ValueError("Invalid weighted train table",
                             response.error_message)

    def update(
        self,
        source_table: SourceTable,
        train_table_mod: TrainingTableSpec,
        validate: bool = True,
    ) -> Self:
        r"""Sets the `source_table` as the modified training table.

        .. note::
            The only allowed modification is the addition of weight column
            Any other modification might lead to unintentded ERRORS while
            training.
            Further negative/NA weight values are not supported.

        The custom training table is ingested during trainer.fit()
        and is used as the training table.

        Args:
            source_table (SourceTable): The source table to be used as the
                modified training table.
            table_mod_spec (TrainTableSpec): The modification specification.
            validate (bool): Whether to validate the modified training table.
                This can be slow for large tables.

        """
        if isinstance(source_table.connector, S3Connector):
            # Special handling for s3 as `source_table._to_api_source_table`
            # concatenates root_dir and file name. But the backend expects
            # these to be separate.
            source_table_type = self._to_s3_api_source_table(source_table)
        else:
            source_table_type = source_table._to_api_source_table()
        if validate:
            self.validate_custom_table(source_table_type, train_table_mod)
        self._custom_train_table = CustomTrainingTable(
            source_table=source_table_type, table_mod_spec=train_table_mod,
            validated=validate)
        return self


# Training Table Future #######################################################


class TrainingTableJob(KumoProgressFuture[TrainingTable]):
    r"""A representation of an ongoing training table generation job in the
    Kumo platform.

    .. code-block:: python

        import kumoai

        # See `PredictiveQuery` documentation:
        pquery = kumoai.PredictiveQuery(...)

        # If a training table is generated in nonblocking mode, the response
        # will be of type `TrainingTableJob`:
        training_table_job = pquery.generate_training_table(non_blocking=True)

        # You can also construct a `TrainingTableJob` from a job ID, e.g.
        # one that is present in the Kumo Jobs page:
        training_table_job = kumoai.TrainingTableJob("trainingjob-...")

        # Get the status of the job:
        print(training_table_job.status())

        # Attach to the job, and poll progress updates:
        training_table_job.attach()

        # Cancel the job:
        training_table_job.cancel()

        # Wait for the job to complete, and return a `TrainingTable`:
        training_table_job.result()

    Args:
        job_id: ID of the training table generation job.
    """
    def __init__(
        self,
        job_id: GenerateTrainTableJobID,
    ) -> None:
        self.job_id = job_id
        # A training table holds a reference to the future that tracks the
        # execution of its generation.
        self._fut: Future[TrainingTable] = create_future(_poll(job_id))

    @property
    def id(self) -> GenerateTrainTableJobID:
        r"""The unique ID of this training table generation process."""
        return self.job_id

    @override
    def result(self) -> TrainingTable:
        return self._fut.result()

    @override
    def future(self) -> Future[TrainingTable]:
        return self._fut

    def status(self) -> JobStatusReport:
        r"""Returns the status of a running training table generation job."""
        return _get_status(self.job_id)

    @override
    def _attach_internal(self, interval_s: float = 20.0) -> TrainingTable:
        assert interval_s >= 4.0
        print(f"Attaching to training table generation job {self.job_id}. "
              f"Tracking this job in the Kumo UI is coming soon. To detach "
              f"from this job, please enter Ctrl+C (the job will continue to "
              f"run, and you can re-attach anytime).")
        api = global_state.client.generate_train_table_job_api

        def _get_progress() -> Optional[Tuple[int, int]]:
            progress = api.get_progress(self.job_id)
            if len(progress) == 0:
                return None
            expected_iter = progress['num_expected_iterations']
            completed_iter = progress['num_finished_iterations']
            return (expected_iter, completed_iter)

        # Print progress bar:
        print("Training table generation is in progress. If your task is "
              "temporal, progress per timeframe will be loaded shortly.")

        # Wait for either timeframes to become available, or the job is done:
        progress = _get_progress()
        while progress is None:
            progress = _get_progress()
            # Not a temporal task, and it's done:
            if self.status().status.is_terminal:
                return self.result()
            time.sleep(interval_s)

        # Wait for timeframes to become available:
        progress = _get_progress()
        assert progress is not None
        total, prog = progress
        pbar = tqdm(total=total, unit="timeframe",
                    desc="Generating Training Table")
        pbar.update(pbar.n - prog)
        while not self.done():
            progress = _get_progress()
            assert progress is not None
            total, prog = progress
            pbar.reset(total)
            pbar.update(prog)
            time.sleep(interval_s)
        pbar.update(pbar.total)
        pbar.close()

        # Future is done:
        return self.result()

    def delete_tags(self, tags: List[str]) -> bool:
        r"""Removes the tags from the job.

        Args:
            tags (List[str]): The tags to remove.
        """
        api = global_state.client.generate_train_table_job_api
        return api.delete_tags(self.job_id, tags)

    def update_tags(self, tags: Mapping[str, Optional[str]]) -> bool:
        r"""Updates the tags of the job.

        Args:
            tags (Mapping[str, Optional[str]]): The tags to update.
                Note that the value 'none' will remove the tag. If the tag is
                not present, it will be added.
        """
        api = global_state.client.generate_train_table_job_api
        return api.update_tags(self.job_id, tags)

    def cancel(self) -> None:
        r"""Cancels a running training table generation job, and raises an
        error if cancellation failed.
        """
        api = global_state.client.generate_train_table_job_api
        return api.cancel(self.job_id)


def _get_status(job_id: str) -> JobStatusReport:
    api = global_state.client.generate_train_table_job_api
    resource: GenerateTrainTableJobResource = api.get(job_id)
    return resource.job_status_report


async def _poll(job_id: str) -> TrainingTable:
    # TODO(manan): make asynchronous natively with aiohttp:
    status = _get_status(job_id).status
    while not status.is_terminal:
        await asyncio.sleep(_DEFAULT_INTERVAL_S)
        status = _get_status(job_id).status

    if status != JobStatus.DONE:
        api = global_state.client.generate_train_table_job_api
        error_details = api.get_job_error(job_id)
        error_str = pretty_print_error_details(error_details)
        raise RuntimeError(
            f"Training table generation for job {job_id} failed with "
            f"job status {status}. Encountered below error(s):"
            f'{error_str}')

    return TrainingTable(job_id)



from .job import ArtifactExportJob, ArtifactExportResult
from .config import OutputConfig, TrainingTableExportConfig

__all__ = [
    "ArtifactExportJob",
    "ArtifactExportResult",
    "OutputConfig",
    "TrainingTableExportConfig",
]



from collections import defaultdict

import pandas as pd
from kumoapi.typing import Dtype, Stype

MAX_CAT = 500


def contains_multicategorical(
    ser: pd.Series,
    column_name: str,
    dtype: Dtype,
) -> bool:

    if not Stype.multicategorical.supports_dtype(dtype):
        return False

    if dtype == Dtype.stringlist:
        return True

    ser = ser.sample(n=1000, random_state=42) if len(ser) > 1000 else ser

    num_unique: int = 0
    if dtype == Dtype.string:
        text = '\n'.join(ser.fillna(''))

        white_list = {';', ':', '|', '\t'}
        candidates: dict[str, int] = defaultdict(int)
        for char in text:
            if char in white_list:
                candidates[char] += 1

        if len(candidates) == 0:
            return False

        num_unique = ser.nunique()

        sep = max(candidates, key=candidates.get)  # type: ignore
        ser = ser.str.split(sep)

    num_unique_multi = ser.explode().nunique()

    if dtype.is_list():
        return num_unique_multi <= MAX_CAT

    return num_unique > 1.5 * num_unique_multi and num_unique_multi <= MAX_CAT



import logging
from typing import (
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    overload,
)

from kumoapi.jobs import (
    BatchPredictionOptions,
    BatchPredictionRequest,
    JobStatus,
    MetadataField,
    PredictionOutputConfig,
    TrainingJobRequest,
    TrainingJobResource,
)
from kumoapi.model_plan import ModelPlan

from kumoai import global_state
from kumoai.artifact_export.config import OutputConfig
from kumoai.client.jobs import (
    GeneratePredictionTableJobID,
    TrainingJobAPI,
    TrainingJobID,
)
from kumoai.connector.base import Connector
from kumoai.connector.s3_connector import S3URI
from kumoai.databricks import to_db_table_name
from kumoai.graph import Graph
from kumoai.pquery.prediction_table import PredictionTable, PredictionTableJob
from kumoai.pquery.training_table import TrainingTable, TrainingTableJob
from kumoai.trainer.job import (
    BatchPredictionJob,
    BatchPredictionJobResult,
    TrainingJob,
    TrainingJobResult,
)
from kumoai.trainer.util import (
    build_prediction_output_config,
    validate_output_arguments,
)

logger = logging.getLogger(__name__)


class Trainer:
    r"""A trainer supports creating a Kumo machine learning model on a
    :class:`~kumoai.pquery.PredictiveQuery`. It is primarily oriented around
    two methods: :meth:`~kumoai.trainer.Trainer.fit`, which accepts a
    :class:`~kumoai.graph.Graph` and :class:`~kumoai.pquery.TrainingTable` and
    produces a :class:`~kumoai.trainer.TrainingJobResult`, and
    :meth:`~kumoai.trainer.Trainer.predict`, which accepts a
    :class:`~kumoai.graph.Graph` and :class:`~kumoai.pquery.PredictionTable`
    and produces a :class:`~kumoai.trainer.BatchPredictionJobResult`.

    A :class:`~kumoai.trainer.Trainer` can also be loaded from a training job,
    with :meth:`~kumoai.trainer.Trainer.load`.

    .. code-block:: python

        import kumoai

        # See `Graph` and `PredictiveQuery` documentation:
        graph = kumoai.Graph(...)
        pquery = kumoai.PredictiveQuery(graph=graph, query=...)

        # Create a `Trainer` object, using a suggested model plan given the
        # predictive query:
        model_plan = pquery.suggest_model_plan()
        trainer = kumoai.Trainer(model_plan)

        # Create a training table from the predictive query:
        training_table = pquery.generate_training_table()

        # Fit a model:
        training_job = trainer.fit(
            graph = graph,
            training_table = training_table,
        )

        # Create a prediction table from the predictive query:
        prediction_table = pquery.generate_prediction_table()

        # Generate predictions:
        prediction_job = trainer.predict(
            graph = graph,
            prediction_table = prediction_table,
            # other arguments here...
        )

        # Load a trained query to generate predictions:
        pquery = kumoai.PredictiveQuery.load_from_training_job("trainingjob-...")
        trainer = kumoai.Trainer.load("trainingjob-...")
        prediction_job = trainer.predict(
            pquery.graph,
            pquery.generate_prediction_table(),
        )

    Args:
        model_plan: A model plan that the trainer should follow when fitting a
            Kumo model to a predictive query. This model plan can either be
            generated from a predictive query, with
            :meth:`~kumoai.pquery.PredictiveQuery.suggest_model_plan`, or can
            be manually specified.
    """  # noqa: E501

    def __init__(self, model_plan: ModelPlan) -> None:
        self._model_plan: Optional[ModelPlan] = model_plan

        # Cached from backend:
        self._training_job_id: Optional[TrainingJobID] = None

    @property
    def model_plan(self) -> Optional[ModelPlan]:
        return self._model_plan

    @model_plan.setter
    def model_plan(self, model_plan: ModelPlan) -> None:
        self._model_plan = model_plan

    # Metadata ################################################################

    @property
    def is_trained(self) -> bool:
        r"""Returns ``True`` if this trainer instance has successfully been
        trained (and is therefore ready for prediction); ``False`` otherwise.
        """
        if not self._training_job_id:
            return False
        try:
            api = global_state.client.training_job_api
            res: TrainingJobResource = api.get(self._training_job_id)
        except Exception as e:  # noqa
            logger.exception(
                "Failed to fetch training status for training job with ID %s",
                self._training_job_id, exc_info=e)
            return False
        return res.job_status_report.status == JobStatus.DONE

    # Fit / predict ###########################################################

    @overload
    def fit(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
    ) -> TrainingJobResult:
        pass

    @overload
    def fit(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
        *,
        non_blocking: Literal[False],
    ) -> TrainingJobResult:
        pass

    @overload
    def fit(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
        *,
        non_blocking: Literal[True],
    ) -> TrainingJob:
        pass

    @overload
    def fit(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
        *,
        non_blocking: bool,
    ) -> Union[TrainingJob, TrainingJobResult]:
        pass

    def fit(
        self,
        graph: Graph,
        train_table: Union[TrainingTable, TrainingTableJob],
        *,
        non_blocking: bool = False,
        custom_tags: Mapping[str, str] = {},
    ) -> Union[TrainingJob, TrainingJobResult]:
        r"""Fits a model to the specified graph and training table, with the
        strategy defined by this :class:`Trainer`'s :obj:`model_plan`.

        Args:
            graph: The :class:`~kumoai.graph.Graph` object that represents the
                tables and relationships that Kumo will learn from.
            train_table: The :class:`~kumoai.pquery.TrainingTable`, or
                in-progress :class:`~kumoai.pquery.TrainingTableJob`, that
                represents the training data produced by a
                :class:`~kumoai.pquery.PredictiveQuery` on :obj:`graph`.
            non_blocking: Whether this operation should return immediately
                after launching the training job, or await completion of the
                training job.
            custom_tags: Additional, customer defined k-v tags to be associated
                with the job to be launched. Job tags are useful for grouping
                and searching jobs.

        Returns:
            Union[TrainingJobResult, TrainingJob]:
                If ``non_blocking=False``, returns a training job object. If
                ``non_blocking=True``, returns a training job future object.
        """
        # TODO(manan, siyang): remove soon:
        job_id = train_table.job_id
        assert job_id is not None

        train_table_job_api = global_state.client.generate_train_table_job_api
        pq_id = train_table_job_api.get(job_id).config.pquery_id
        assert pq_id is not None

        custom_table = None
        if isinstance(train_table, TrainingTable):
            custom_table = train_table._custom_train_table

        # NOTE the backend implementation currently handles sequentialization
        # between a training table future and a training job; that is, if the
        # training table future is still executing, the backend will wait on
        # the job ID completion before executing a training job. This preserves
        # semantics for both futures, ensures that Kumo works as expected if
        # used only via REST API, and allows us to avoid chaining calllbacks
        # in an ugly way here:
        api = global_state.client.training_job_api
        self._training_job_id = api.create(
            TrainingJobRequest(
                dict(custom_tags),
                pquery_id=pq_id,
                model_plan=self._model_plan,
                graph_snapshot_id=graph.snapshot(non_blocking=non_blocking),
                train_table_job_id=job_id,
                custom_train_table=custom_table,
            ))

        out = TrainingJob(job_id=self._training_job_id)
        if non_blocking:
            return out
        return out.attach()

    def predict(
        self,
        graph: Graph,
        prediction_table: Union[PredictionTable, PredictionTableJob],
        output_types: Optional[Set[str]] = None,
        output_connector: Optional[Connector] = None,
        output_table_name: Optional[Union[str, Tuple]] = None,
        output_metadata_fields: Optional[List[MetadataField]] = None,
        output_config: Optional[OutputConfig] = None,
        training_job_id: Optional[TrainingJobID] = None,
        binary_classification_threshold: Optional[float] = None,
        num_classes_to_return: Optional[int] = None,
        num_workers: int = 1,
        non_blocking: bool = False,
        custom_tags: Mapping[str, str] = {},
    ) -> Union[BatchPredictionJob, BatchPredictionJobResult]:
        """Using the trained model specified by :obj:`training_job_id` (or
        the model corresponding to the last invocation of
        :meth:`~kumoai.trainer.Trainer.fit`, if not present), predicts the
        future values of the entities in :obj:`prediction_table` leveraging
        current information from :obj:`graph`.

        Args:
            graph: The :class:`~kumoai.graph.Graph` object that represents the
                tables and relationships that Kumo will use to make
                predictions.
            prediction_table: The :class:`~kumoai.pquery.PredictionTable`, or
                in-progress :class:`~kumoai.pquery.PredictionTableJob`, that
                represents the prediction data produced by a
                :class:`~kumoai.pquery.PredictiveQuery` on :obj:`graph`. This
                table may also be custom-generated.
            output_config: Output configuration defining properties of the
                generated batch prediction outputs. This includes:

                - output_types: The types of outputs that should be produced by
                  the prediction job. Can include either ``'predictions'``,
                  ``'embeddings'``, or both.
                - output_connector: The output data source that Kumo should
                  write batch predictions to, if it is None, produce
                  local download output only.
                - output_table_name: The name of the table in the output data
                  source that Kumo should write batch predictions to. In the
                  case of a Databricks connector, this should be a tuple of
                  two strings, the schema name and the output prediction
                  table name.
                - output_metadata_fields: Any additional metadata fields to
                  include as new columns in the produced ``'predictions'``
                  output. Currently, allowed options are ``JOB_TIMESTAMP``
                  and ``ANCHOR_TIMESTAMP``.
                - connector_specific_config: Custom connector specific output
                  configuration, such as whether to append or overwrite
                  existing tables.

            output_types: *(Deprecated)* The types of outputs that should be
                produced by the prediction job. Can include either
                ``'predictions'``, ``'embeddings'``, or both. Use
                :obj:`output_config` instead.
            output_connector: *(Deprecated)* The output data source that Kumo
                should write batch predictions to, if it is None, produce local
                download output only. Use :obj:`output_config` instead.
            output_table_name: *(Deprecated)* The name of the table in the
                output data source that Kumo should write batch predictions to.
                In the case of a Databricks connector, this should be a tuple
                of two strings: the schema name and the output prediction
                table name. Use :obj:`output_config` instead.
            output_metadata_fields: *(Deprecated)* Any additional metadata
                fields to include as new columns in the produced
                ``'predictions'`` output. Currently, allowed options are
                ``JOB_TIMESTAMP`` and ``ANCHOR_TIMESTAMP``. Use
                :obj:`output_config` instead.
            training_job_id: The job ID of the training job whose model will be
                used for making predictions.
            binary_classification_threshold: If this model corresponds to
                a binary classification task, the threshold for which higher
                values correspond to ``1``, and lower values correspond to
                ``0``.
            num_classes_to_return: If this model corresponds to a ranking task,
                the number of classes to return in the prediction output.
            num_workers: Number of workers to use when generating batch
                predictions. When set to a value greater than 1, the prediction
                table is partitioned into smaller parts and processed in
                parallel. The default is 1, which implies sequential inference
                over the prediction table.
            non_blocking: Whether this operation should return immediately
                after launching the batch prediction job, or await
                completion of the batch prediction job.
            custom_tags: Additional, customer defined k-v tags to be associated
                with the job to be launched. Job tags are useful for grouping
                and searching jobs.

        Returns:
            Union[BatchPredictionJob, BatchPredictionJobResult]:
                If ``non_blocking=False``, returns a training job object. If
                ``non_blocking=True``, returns a training job future object.
        """
        if (output_types is not None or output_connector is not None
                or output_table_name is not None
                or output_metadata_fields is not None):
            raise ValueError(
                'Specifying output_types, output_connector, '
                'output_metadata_fields '
                'and output_table_name as direct inputs to predict() is '
                'deprecated. Please use output_config to specify these '
                'parameters.')
        assert output_config is not None
        output_table_name = to_db_table_name(output_config.output_table_name)
        validate_output_arguments(
            output_config.output_types,
            output_config.output_connector,
            output_table_name,
        )

        # Create outputs:
        outputs: List[PredictionOutputConfig] = []
        for output_type in output_config.output_types:
            if output_config.output_connector is None:
                # Predictions are generated to the Kumo dataplane, and can
                # only be exported via the UI for now:
                pass
            else:
                outputs.append(
                    build_prediction_output_config(
                        output_type,
                        output_config.output_connector,
                        output_table_name,
                        output_config.output_metadata_fields,
                        output_config,
                    ))

        training_job_id = training_job_id or self._training_job_id
        if training_job_id is None:
            raise ValueError(
                "Cannot run batch prediction without a specified or saved "
                "training job ID. Please either call `fit(...)` or pass a "
                "job ID of a completed training job to proceed.")

        pred_table_job_id: Optional[GeneratePredictionTableJobID] = \
            prediction_table.job_id
        pred_table_data_path = None
        if pred_table_job_id is None:
            assert isinstance(prediction_table, PredictionTable)
            if isinstance(prediction_table.table_data_uri, S3URI):
                pred_table_data_path = prediction_table.table_data_uri.uri
            else:
                pred_table_data_path = prediction_table.table_data_uri

        api = global_state.client.batch_prediction_job_api

        job_id, response = api.maybe_create(
            BatchPredictionRequest(
                dict(custom_tags),
                model_training_job_id=training_job_id,
                predict_options=BatchPredictionOptions(
                    binary_classification_threshold=(
                        binary_classification_threshold),
                    num_classes_to_return=num_classes_to_return,
                    num_workers=num_workers,
                ),
                outputs=outputs,
                graph_snapshot_id=graph.snapshot(non_blocking=non_blocking),
                pred_table_job_id=pred_table_job_id,
                pred_table_path=pred_table_data_path,
            ))

        message = response.message()
        if not response.ok:
            raise RuntimeError(f"Prediction failed. {message}")
        elif not response.empty():
            logger.warning("Prediction produced the following warnings: %s",
                           message)
        assert job_id is not None

        self._batch_prediction_job_id = job_id
        out = BatchPredictionJob(job_id=self._batch_prediction_job_id)
        if non_blocking:
            return out
        return out.attach()

    # Persistence #############################################################

    @classmethod
    def _load_from_job(cls, job: TrainingJobResource) -> 'Trainer':
        trainer = cls(job.config.model_plan)
        trainer._training_job_id = job.job_id
        return trainer

    @classmethod
    def load(cls, job_id: TrainingJobID) -> 'Trainer':
        r"""Creates a :class:`~kumoai.trainer.Trainer` instance from a training
        job ID.
        """
        api: TrainingJobAPI = global_state.client.training_job_api
        job = api.get(job_id)
        return cls._load_from_job(job)

    # TODO(siyang): load trainer by searching training job via tags.
    @classmethod
    def load_from_tags(cls, tags: Mapping[str, str]) -> 'Trainer':
        r"""Creates a :class:`~kumoai.trainer.Trainer` instance from a set of
        job tags. If multiple jobs match the list of tags, only one will be
        selected.
        """
        api = global_state.client.training_job_api
        jobs = api.list(limit=1, additional_tags=tags)
        if not jobs:
            raise RuntimeError(f'No successful training job found for {tags}')
        assert len(jobs) == 1
        return cls._load_from_job(jobs[0])



import logging
from typing import List, Optional

from kumoapi.data_source import DataSourceType, FileConnectorResourceConfig
from kumoapi.source_table import (
    S3SourceTableRequest,
    SourceTableConfigRequest,
    SourceTableConfigResponse,
    SourceTableListRequest,
    SourceTableValidateRequest,
    SourceTableValidateResponse,
)
from typing_extensions import Self, override

from kumoai import global_state
from kumoai.connector import Connector
from kumoai.connector.source_table import SourceTable

logger = logging.getLogger(__name__)

_DEFAULT_NAME = 's3_connector'


class S3Connector(Connector):
    r"""Defines a connector to a table stored as a file (or partitioned
    set of files) on the Amazon `S3 <https://aws.amazon.com/s3/>`__ object
    store. Any table behind an S3 bucket accessible by the shared external IAM
    role can be accessed through this connector.

    .. code-block:: python

        import kumoai
        connector = kumoai.S3Connector(root_dir="s3://...")  # an S3 path.

        # List all tables:
        print(connector.table_names())  # Returns: ['articles', 'customers', 'users']

        # Check whether a table is present:
        assert "articles" in connector

        # Fetch a source table (both approaches are equivalent):
        source_table = connector["articles"]
        source_table = connector.table("articles")

    Args:
        root_dir: The root directory of this connector. If provided, the root
            directory is used as a prefix for tables in this connector. If not
            provided, all tables must be specified by their full S3 paths.
    """  # noqa

    def __init__(
        self,
        root_dir: Optional[str] = None,
        _connector_id: Optional[str] = None,
    ) -> None:
        if _connector_id is not None:
            # UI S3Connector, named:
            self._connector_id = _connector_id
            self.root_dir = None
            return

        self._connector_id = _DEFAULT_NAME
        if root_dir is not None:
            # Remove trailing / to be consistent with boto s3
            root_dir = root_dir.rstrip('/')
        self.root_dir = root_dir
        if global_state.is_spcs and root_dir is not None \
                and root_dir.startswith('s3://'):
            raise ValueError(
                "S3 connectors are not supported when running Kumo in "
                "Snowpark container services. Please use a Snowflake "
                "connector instead.")

    @override
    @property
    def name(self) -> str:
        r"""Not supported by :class:`S3Connector`; returns an internal
        specifier.
        """
        return self._connector_id

    @override
    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.S3

    @override
    def _source_table_request(
        self,
        table_names: List[str],
    ) -> S3SourceTableRequest:
        root_dir = self.root_dir
        if not root_dir and self.name == _DEFAULT_NAME:
            # Handle None root directories (table name is a path):
            table_path = S3URI(table_names[0]).validate()
            root_dir = table_path.root_dir
            for i, v in enumerate(table_names):
                uri = S3URI(v)
                if uri.root_dir != root_dir:
                    # TODO(manan): fix
                    raise ValueError(
                        f"Please ensure that all of your tables are behind "
                        f"the same root directory ({root_dir}).")
                table_names[i] = uri.object_name

        connector_id = self.name if self.name != _DEFAULT_NAME else None
        root_dir = root_dir if self.name == _DEFAULT_NAME else ""

        # TODO(manan): file type?
        return S3SourceTableRequest(
            s3_root_dir=root_dir,
            connector_id=connector_id,
            table_names=table_names,
        )

    @override
    def table(self, name: str) -> SourceTable:
        r"""Returns a :class:`~kumoai.connector.SourceTable` object
        corresponding to a source table on Amazon S3.

        Args:
            name: The name of the table on S3. If :obj:`root_dir` is provided,
                the path will be specified as :obj:`root_dir/name`. If
                :obj:`root_dir` is not provided, the name should be the full
                path (e.g. starting with ``s3://``).

        Raises:
            :class:`ValueError`: if ``name`` does not exist in the backing
                connector.
        """
        # NOTE only overridden for documentation purposes.
        return super().table(name)

    @override
    def _list_tables(self) -> List[str]:
        connector_id = self.name if self.name != _DEFAULT_NAME else None
        root_dir = self.root_dir if self.name == _DEFAULT_NAME else None
        if root_dir is None and connector_id is None:
            raise ValueError(
                "Listing tables without a specified root directory is not "
                "supported. Please specify a root directory to continue; "
                "alternatively, please access individual tables with their "
                "full S3 paths.")

        req = SourceTableListRequest(connector_id=connector_id,
                                     root_dir=root_dir,
                                     source_type=DataSourceType.S3)
        return global_state.client.source_table_api.list_tables(req)

    @override
    def _get_table_config(self, table_name: str) -> SourceTableConfigResponse:
        root_dir = self.root_dir
        if not root_dir and self.name == _DEFAULT_NAME:
            # Handle None root directories (table name is a path):
            table_path = S3URI(table_name).validate()
            root_dir = table_path.root_dir
            table_name = table_path.object_name

        connector_id = self.name if self.name != _DEFAULT_NAME else None
        root_dir = root_dir if self.name == _DEFAULT_NAME else None

        req = SourceTableConfigRequest(
            connector_id=connector_id,
            root_dir=root_dir,
            table_name=table_name,
            source_type=self.source_type,
        )
        return global_state.client.source_table_api.get_table_config(req)

    @override
    def _validate_table(self, table_name: str) -> SourceTableValidateResponse:
        if table_name in self._validated_tables:
            return SourceTableValidateResponse(is_valid=True, msg='')

        if self.name == _DEFAULT_NAME:
            # For S3 connector without name, pass root_dir to validate table
            req = SourceTableValidateRequest(
                connector_id=None,
                table_name=table_name,
                source_type=self.source_type,
                root_dir=self.root_dir,
            )
        else:
            req = SourceTableValidateRequest(connector_id=self.name,
                                             table_name=table_name,
                                             source_type=self.source_type)

        ret = global_state.client.source_table_api.validate_table(req)
        # Cache the result for the whole session.
        if ret.is_valid:
            self._validated_tables.add(table_name)
        return ret

    # Class properties ########################################################

    @classmethod
    def get_by_name(cls, name: str) -> Self:
        r"""Returns an instance of a named S3 Connector, created in the Kumo UI.

        .. note::
            Named S3 connectors are read-only: if you would like to modify the
            root directory, please do so from the UI.

        Args:
            name: The name of the existing connector.

        Example:
            >>> import kumoai
            >>> connector = kumoai.S3Connector.get_by_name("name")  # doctest: +SKIP # noqa: E501
        """
        api = global_state.client.connector_api
        resp = api.get(name)
        if resp is None:
            raise ValueError(
                f"There does not exist an existing stored connector with name "
                f"{name}.")
        config = resp.config
        assert isinstance(config, FileConnectorResourceConfig)
        return cls(
            root_dir=None,
            _connector_id=config.name,
        )

    def __repr__(self) -> str:
        if self.name != _DEFAULT_NAME:
            return f'{self.__class__.__name__}(name={self.name})'
        root_dir_name = f"\"{self.root_dir}\"" if self.root_dir else "None"
        return f'{self.__class__.__name__}(root_dir={root_dir_name})'


class S3URI:
    r"""A utility class to parse and navigate S3 URIs."""
    def __init__(self, uri: str):
        self.uri: str = uri
        if uri.endswith('/'):  # remove trailing slash
            self.uri = uri[:-1]

    @property
    def is_valid(self) -> bool:
        # TODO(zeyuan): For SPCS, the path can be a local filesystem path
        # For train/pred table.
        if global_state.is_spcs:
            return True
        # TODO(manan): implement more checks...
        return self.uri.startswith("s3://")

    def validate(self) -> Self:
        if not self.is_valid:
            raise ValueError(f"Path {self.uri} is not a valid S3 URI.")
        return self

    @property
    def root_dir(self) -> str:
        self.validate()
        return self.uri.rsplit('/', 1)[0]

    @property
    def object_name(self) -> str:
        self.validate()
        return self.uri.rsplit('/', 1)[1]

    # Class properties ########################################################

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}('
                f'uri={self.uri}, valid={self.is_valid})')
